import re
from pathlib import Path

from ..common import XMLReader, read_xml, save_xml
from ..pdf import TITLE_TAGS, Page
from ..pdf import decode as decode_pdf
from .toc_levels import Ref2Level, analyse_title_levels, analyse_toc_levels
from .toc_pages import PageRef, find_toc_pages
from .types import Toc, TocInfo
from .types import decode as decode_toc
from .types import encode as encode_toc

_TITLE_HEAD_REGX = re.compile(r"^\s*#{1,6}\s*")


def analyse_toc(pages_path: Path, toc_path: Path, toc_assumed: bool) -> TocInfo:
    if toc_path.exists():
        return decode_toc(read_xml(toc_path))

    toc_path.parent.mkdir(parents=True, exist_ok=True)
    toc_info = _do_analyse_toc(pages_path, toc_assumed)
    save_xml(encode_toc(toc_info), toc_path)

    return toc_info


def _do_analyse_toc(pages_path: Path, toc_assumed: bool) -> TocInfo:
    pages: XMLReader[Page] = XMLReader(
        prefix="page",
        dir_path=pages_path,
        decode=decode_pdf,
    )
    ref2level: Ref2Level
    toc_page_indexes: list[int] = []
    toc_pages: list[PageRef] = []
    if toc_assumed:
        toc_pages = find_toc_pages(
            iter_titles=lambda: (
                list(
                    (layout.order, _TITLE_HEAD_REGX.sub("", layout.text))
                    for layout in page.body_layouts
                    if layout.ref in TITLE_TAGS
                )
                for page in pages.read()
            ),
            iter_page_bodies=lambda: (
                "".join(layout.text for layout in page.body_layouts)
                for page in pages.read()
            ),
        )

    if toc_pages:
        ref2level = analyse_toc_levels(
            pages=pages,
            pages_path=pages_path,
            toc_pages=toc_pages,
        )
        toc_page_indexes.extend(ref.page_index for ref in toc_pages)
    else:
        ref2level = analyse_title_levels(pages)

    return TocInfo(
        content=_structure_toc_by_levels(ref2level),
        page_indexes=sorted(toc_page_indexes),
    )


def _structure_toc_by_levels(ref2level: Ref2Level) -> list[Toc]:
    # 虚拟根节点
    root = Toc(
        id=-1,
        page_index=-1,
        order=-1,
        level=-1,
        children=[],
    )
    next_id: int = 1
    stack: list[Toc] = [root]

    for (page_index, order), level in sorted(ref2level.items(), key=lambda x: x[0]):
        toc = Toc(
            id=next_id,
            page_index=page_index,
            order=order,
            level=level,
            children=[],
        )
        next_id += 1
        while stack and stack[-1].level >= level:
            stack.pop()

        if not stack:
            break  # 防御性

        stack[-1].children.append(toc)
        stack.append(toc)

    return root.children



from pathlib import Path
from shutil import copy2
from typing import Callable, Generator, Iterable

from ...expression import ExpressionKind, to_markdown_string
from ...pdf import TITLE_TAGS
from ...sequence import (
    AssetLayout,
    BlockMember,
    InlineExpression,
    ParagraphLayout,
    Reference,
    RefIdMap,
)
from ..paragraph import render_markdown_paragraph

_MAX_TOC_LEVELS = 3
_MAX_TITLE_LEVELS = 6


def render_layouts(
    layouts: Iterable[ParagraphLayout | AssetLayout],
    assets_path: Path,
    output_assets_path: Path,
    asset_ref_path: Path,
    toc_level: int,
    ref_id_to_number: RefIdMap | None = None,
) -> Generator[str, None, None]:
    is_first_layout = True
    toc_level = min(toc_level, _MAX_TOC_LEVELS - 1)

    for layout in layouts:
        if is_first_layout:
            is_first_layout = False
        else:
            yield "\n\n"
        if isinstance(layout, AssetLayout):
            yield from _render_asset(
                asset=layout,
                assets_path=assets_path,
                output_assets_path=output_assets_path,
                asset_ref_path=asset_ref_path,
                ref_id_to_number=ref_id_to_number,
            )
        elif isinstance(layout, ParagraphLayout):
            yield from render_paragraph(
                paragraph=layout,
                toc_level=toc_level,
                ref_id_to_number=ref_id_to_number,
            )


def render_paragraph(
    paragraph: ParagraphLayout, toc_level: int, ref_id_to_number: RefIdMap | None = None
) -> Generator[str, None, None]:
    if paragraph.level >= 0 and paragraph.ref in TITLE_TAGS:
        level = min(toc_level + paragraph.level, _MAX_TITLE_LEVELS)
        for _ in range(level + 1):  # level 0 对应 1 个 #
            yield "#"
        yield " "

    def render_member(part: BlockMember | str) -> Generator[str, None, None]:
        if isinstance(part, str):
            yield to_markdown_string(
                kind=ExpressionKind.TEXT,
                content=part,
            )
        elif isinstance(part, InlineExpression):
            latex_content = part.content.strip()
            if latex_content:
                yield to_markdown_string(
                    kind=part.kind,
                    content=latex_content,
                )
        elif ref_id_to_number and isinstance(part, Reference):
            ref_number = ref_id_to_number.get(part.id, 1)
            yield "[^"
            yield str(ref_number)
            yield "]"

    for block in paragraph.blocks:
        yield from render_markdown_paragraph(
            children=block.content,
            render_payload=render_member,
        )


_MemberRender = Callable[[BlockMember | str], Iterable[str]]


def _render_asset(
    asset: AssetLayout,
    assets_path: Path,
    output_assets_path: Path,
    asset_ref_path: Path,
    ref_id_to_number: RefIdMap | None = None,
) -> Generator[str, None, None]:
    def render_member(part: BlockMember | str) -> Generator[str, None, None]:
        if isinstance(part, str):
            yield to_markdown_string(
                kind=ExpressionKind.TEXT,
                content=part,
            )
        elif isinstance(part, InlineExpression):
            latex_content = part.content.strip()
            if latex_content:
                yield to_markdown_string(
                    kind=part.kind,
                    content=latex_content,
                )
        elif ref_id_to_number and isinstance(part, Reference):
            ref_number = ref_id_to_number.get(part.id, 1)
            yield "[^"
            yield str(ref_number)
            yield "]"

    has_content = False

    if asset.title:
        title_str = "".join(
            render_markdown_paragraph(
                children=asset.title,
                render_payload=render_member,
            )
        ).strip()
        if title_str:
            yield title_str
            has_content = True

    yield from _render_asset_content(
        asset=asset,
        assets_path=assets_path,
        output_assets_path=output_assets_path,
        asset_ref_path=asset_ref_path,
        render_member=render_member,
        has_content_before=has_content,
    )
    if asset.ref in ("equation", "table"):
        if asset.content:
            has_content = True
    elif asset.ref == "image":
        if asset.hash:
            has_content = True

    if asset.caption:
        caption_str = "".join(
            render_markdown_paragraph(
                children=asset.caption,
                render_payload=render_member,
            )
        ).strip()
        if caption_str:
            if has_content:
                yield "\n\n"
            yield caption_str


def _render_asset_content(
    asset: AssetLayout,
    assets_path: Path,
    output_assets_path: Path,
    asset_ref_path: Path,
    render_member: _MemberRender,
    has_content_before: bool,
) -> Generator[str, None, None]:
    if asset.ref == "equation":
        content_str = "".join(
            render_markdown_paragraph(
                children=asset.content,
                render_payload=render_member,
            )
        )
        latex_content = content_str.strip()
        if latex_content:
            if has_content_before:
                yield "\n\n"
            yield to_markdown_string(
                kind=ExpressionKind.DISPLAY_BRACKET,
                content=latex_content,
            )

    elif asset.ref == "table":
        if asset.content:
            if has_content_before:
                yield "\n\n"
            yield from render_markdown_paragraph(
                children=asset.content,
                render_payload=render_member,
            )

    elif asset.ref == "image":
        yield from _render_image(
            asset=asset,
            assets_path=assets_path,
            output_assets_path=output_assets_path,
            asset_ref_path=asset_ref_path,
            has_content_before=has_content_before,
        )


def _render_image(
    asset: AssetLayout,
    assets_path: Path,
    output_assets_path: Path,
    asset_ref_path: Path,
    has_content_before: bool,
) -> Generator[str, None, None]:
    # 渲染图片
    if asset.hash is None:
        return

    source_file = assets_path / f"{asset.hash}.png"
    if not source_file.exists():
        return

    target_file = output_assets_path / f"{asset.hash}.png"
    if not target_file.exists():
        copy2(source_file, target_file)

    if asset_ref_path.is_absolute():
        image_path = target_file
    else:
        image_path = asset_ref_path / f"{asset.hash}.png"

    # 使用 POSIX 风格路径(markdown 标准)
    image_path_str = str(image_path).replace("\\", "/")

    # 图片的 alt 保持空
    if has_content_before:
        yield "\n\n"
    yield f"![]({image_path_str})"



from os import PathLike
from pathlib import Path
from typing import Generator

from PIL.Image import Image

from ..error import PDFError
from .handler import DefaultPDFHandler, PDFDocument, PDFHandler


def pdf_pages_count(
    pdf_path: PathLike | str,
    pdf_handler: PDFHandler | None = None,
) -> int:
    if pdf_handler is None:
        pdf_handler = DefaultPDFHandler()
    document: PDFDocument | None = None
    try:
        document = pdf_handler.open(pdf_path=Path(pdf_path))
        return document.pages_count
    except PDFError:
        raise
    except Exception as error:
        raise PDFError("Failed to parse PDF document.", page_index=None) from error
    finally:
        if document is not None:
            document.close()


class PageRefContext:
    def __init__(
        self,
        pdf_path: Path,
        pdf_handler: PDFHandler,
    ) -> None:
        self._pdf_path = pdf_path
        self._pdf_handler: PDFHandler = pdf_handler
        self._document: PDFDocument | None = None

    @property
    def pages_count(self) -> int:
        assert self._document is not None
        return self._document.pages_count

    def __enter__(self) -> "PageRefContext":
        assert self._document is None
        try:
            self._document = self._pdf_handler.open(self._pdf_path)
        except PDFError:
            raise
        except Exception as error:
            raise PDFError("Failed to open PDF document.", page_index=None) from error
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self._document is not None:
            self._document.close()
            self._document = None

    def __iter__(self) -> Generator["PageRef", None, None]:
        assert self._document is not None
        for i in range(self._document.pages_count):
            yield PageRef(
                document=self._document,
                page_index=i + 1,
            )


_PNG_COMPRESSION_RATIO = 0.5  # Conservative estimate for document images
_BYTES_PER_PIXEL = 3  # RGB


class PageRef:
    def __init__(
        self,
        document: PDFDocument,
        page_index: int,
    ) -> None:
        self._document = document
        self._page_index = page_index

    @property
    def page_index(self) -> int:
        return self._page_index

    def render(self, dpi: int, max_image_file_size: int | None = None) -> Image:
        try:
            if max_image_file_size is not None:
                width_inch, height_inch = self._document.page_size(self._page_index)
                max_dpi = round(
                    self._dpi_with_size(
                        file_size=max_image_file_size,
                        width_inch=width_inch,
                        height_inch=height_inch,
                    )
                )
                dpi = min(dpi, max_dpi)

            return self._document.render_page(
                page_index=self._page_index,
                dpi=dpi,
            )
        except PDFError as error:
            error.page_index = self._page_index
            raise error

        except Exception as error:
            raise PDFError(
                f"Failed to render page {self._page_index}.",
                page_index=self._page_index,
            ) from error

    def _dpi_with_size(
        self, file_size: int, width_inch: float, height_inch: float
    ) -> float:
        # Formula: file_size = width_px * height_px * bytes_per_pixel * compression_ratio
        # where width_px = width_inch * dpi, height_px = height_inch * dpi
        return (
            file_size
            / (width_inch * height_inch * _BYTES_PER_PIXEL * _PNG_COMPRESSION_RATIO)
        ) ** 0.5



from dataclasses import dataclass
from typing import Generator, Iterable, cast
from xml.etree.ElementTree import Element

from ..common import ASSET_TAGS, AssetRef, indent
from ..expression import ExpressionKind, decode_expression_kind, encode_expression_kind
from ..markdown.paragraph import HTMLTag, flatten
from ..markdown.paragraph import decode as decode_content
from ..markdown.paragraph import encode as encode_content
from .mark import Mark


@dataclass
class Chapter:
    id: int | None
    level: int
    layouts: list["ParagraphLayout | AssetLayout"]


@dataclass
class ParagraphLayout:
    ref: str
    level: int
    blocks: list["BlockLayout"]


@dataclass
class InlineExpression:
    kind: ExpressionKind
    content: str


@dataclass
class Reference:
    page_index: int
    order: int
    mark: str | Mark
    layouts: list["AssetLayout | ParagraphLayout"]

    @property
    def id(self) -> tuple[int, int]:
        return (self.page_index, self.order)


BlockMember = InlineExpression | Reference
RefIdMap = dict[tuple[int, int], int]


@dataclass
class AssetLayout:
    page_index: int
    ref: AssetRef
    det: tuple[int, int, int, int]
    title: list[str | BlockMember | HTMLTag[BlockMember]]
    content: list[str | BlockMember | HTMLTag[BlockMember]]
    caption: list[str | BlockMember | HTMLTag[BlockMember]]
    hash: str | None


@dataclass
class BlockLayout:
    page_index: int
    order: int
    det: tuple[int, int, int, int]
    content: list[str | BlockMember | HTMLTag[BlockMember]]


def search_references_in_chapter(chapter: Chapter) -> Generator[Reference, None, None]:
    seen: set[tuple[int, int]] = set()
    for part in _search_parts_in_chapter(chapter):
        if isinstance(part, Reference):
            ref_id = part.id
            if ref_id not in seen:
                seen.add(ref_id)
                yield part


def references_to_map(references: Iterable[Reference]) -> RefIdMap:
    ref_id_to_number = {}
    for i, ref in enumerate(references, 1):
        ref_id_to_number[ref.id] = i
    return ref_id_to_number


def decode(element: Element) -> Chapter:
    references_el = element.find("references")
    references_map: dict[tuple[int, int], Reference] = {}
    if references_el is not None:
        for ref_el in references_el.findall("ref"):
            reference = _decode_reference(ref_el)
            references_map[reference.id] = reference

    id_attr = element.get("id")
    id_value = int(id_attr) if id_attr is not None else None

    level_attr = element.get("level")
    level = int(level_attr) if level_attr is not None else -1

    body_el = element.find("body")
    if body_el is None:
        raise ValueError("<chapter> missing required <body> element")

    layouts: list[ParagraphLayout | AssetLayout] = []
    for child in list(body_el):
        tag = child.tag
        if tag == "asset":
            layouts.append(_decode_asset(child))
        elif tag == "paragraph":
            layouts.append(_decode_paragraph(child, references_map))

    return Chapter(
        id=id_value,
        level=level,
        layouts=layouts,
    )


def encode(chapter: Chapter) -> Element:
    root = Element("chapter")

    if chapter.id is not None:
        root.set("id", str(chapter.id))

    if chapter.level != -1:
        root.set("level", str(chapter.level))

    body_el = Element("body")
    for layout in chapter.layouts:
        if isinstance(layout, AssetLayout):
            body_el.append(_encode_asset(layout))
        else:
            body_el.append(_encode_paragraph(layout))

    root.append(body_el)
    references = list(search_references_in_chapter(chapter))
    references.sort(key=lambda ref: ref.id)

    if references:
        references_el = Element("references")
        for ref in references:
            references_el.append(_encode_reference(ref))
        root.append(references_el)

    return indent(root)


def _search_parts_in_chapter(chapter: Chapter):
    for layout in chapter.layouts:
        if isinstance(layout, ParagraphLayout):
            for block in layout.blocks:
                yield from flatten(block.content)


def _decode_asset(element: Element) -> AssetLayout:
    ref_attr = element.get("ref")
    if ref_attr is None:
        raise ValueError("<asset> missing required attribute 'ref'")
    if ref_attr not in ASSET_TAGS:
        raise ValueError(
            f"<asset> attribute 'ref' must be one of {ASSET_TAGS}, got: {ref_attr}"
        )
    page_index_attr = element.get("page_index")
    if page_index_attr is None:
        raise ValueError("<asset> missing required attribute 'page_index'")
    try:
        page_index = int(page_index_attr)
    except ValueError as e:
        raise ValueError(
            f"<asset> attribute 'page_index' must be int, got: {page_index_attr}"
        ) from e

    det_str = element.get("det")
    if det_str is None:
        raise ValueError("<asset> missing required attribute 'det'")
    det = _parse_det(det_str, context="<asset>@det")

    hash_value = element.get("hash")

    def decode_block_member(child: Element) -> BlockMember:
        if child.tag == "ref":
            raise ValueError("<asset> cannot contain <ref> elements")
        elif child.tag == "inline_expr":
            kind_attr = child.get("kind")
            if kind_attr is None:
                raise ValueError(
                    "<asset><inline_expr> missing required attribute 'kind'"
                )
            kind = decode_expression_kind(kind_attr)
            expr_text = child.text if child.text is not None else ""
            return InlineExpression(kind=kind, content=expr_text)
        else:
            raise ValueError(f"<asset> contains unknown element: <{child.tag}>")

    title_el = element.find("title")
    title = (
        decode_content(title_el, decode_block_member) if title_el is not None else []
    )

    content_el = element.find("content")
    content = (
        decode_content(content_el, decode_block_member)
        if content_el is not None
        else []
    )

    caption_el = element.find("caption")
    caption = (
        decode_content(caption_el, decode_block_member)
        if caption_el is not None
        else []
    )

    return AssetLayout(
        page_index=page_index,
        ref=cast(AssetRef, ref_attr),
        det=det,
        title=title,
        content=content,
        caption=caption,
        hash=hash_value,
    )


def _encode_asset(layout: AssetLayout) -> Element:
    el = Element("asset")
    el.set("ref", layout.ref)
    el.set("page_index", str(layout.page_index))
    el.set("det", ",".join(map(str, layout.det)))
    if layout.hash is not None:
        el.set("hash", layout.hash)

    if layout.title:
        title_el = Element("title")
        encode_content(
            root=title_el,
            children=layout.title,
            encode_payload=_encode_block_member,
        )
        el.append(title_el)

    if layout.content:
        content_el = Element("content")
        encode_content(
            root=content_el,
            children=layout.content,
            encode_payload=_encode_block_member,
        )
        el.append(content_el)

    if layout.caption:
        caption_el = Element("caption")
        encode_content(
            root=caption_el,
            children=layout.caption,
            encode_payload=_encode_block_member,
        )
        el.append(caption_el)

    return el


def _decode_paragraph(
    element: Element, references_map: dict[tuple[int, int], Reference] | None = None
) -> ParagraphLayout:
    ref_attr = element.get("ref")
    if ref_attr is None:
        raise ValueError("<paragraph> missing required attribute 'ref'")

    level_attr = element.get("level")
    level = int(level_attr) if level_attr is not None else -1

    blocks = _decode_block_elements(
        parent=element,
        context_tag="paragraph",
        references_map=references_map,
    )
    return ParagraphLayout(ref=ref_attr, level=level, blocks=blocks)


def _encode_paragraph(layout: ParagraphLayout) -> Element:
    el = Element("paragraph")
    el.set("ref", layout.ref)
    if layout.level != -1:
        el.set("level", str(layout.level))
    for block in layout.blocks:
        el.append(_encode_block_element(block))
    return el


def _parse_det(det_str: str, context: str) -> tuple[int, int, int, int]:
    try:
        det_list = list(map(int, det_str.split(",")))
    except Exception as e:
        raise ValueError(
            f"{context}: det must be comma-separated integers, got: {det_str}"
        ) from e
    if len(det_list) != 4:
        raise ValueError(f"{context}: det must have 4 values, got {len(det_list)}")
    return (det_list[0], det_list[1], det_list[2], det_list[3])


def _decode_block_elements(
    parent: Element,
    context_tag: str,
    references_map: dict[tuple[int, int], Reference] | None = None,
) -> list[BlockLayout]:
    blocks: list[BlockLayout] = []
    for block_el in parent.findall("block"):
        page_index_attr = block_el.get("page_index")
        if page_index_attr is None:
            raise ValueError(
                f"<{context_tag}><block> missing required attribute 'page_index'"
            )
        try:
            page_index = int(page_index_attr)
        except ValueError as e:
            raise ValueError(
                f"<{context_tag}><block> attribute 'page_index' must be int, got: {page_index_attr}"
            ) from e

        order_attr = block_el.get("order")
        if order_attr is None:
            raise ValueError(
                f"<{context_tag}><block> missing required attribute 'order'"
            )
        try:
            order = int(order_attr)
        except ValueError as e:
            raise ValueError(
                f"<{context_tag}><block> attribute 'order' must be int, got: {order_attr}"
            ) from e

        det_str = block_el.get("det")
        if det_str is None:
            raise ValueError(f"<{context_tag}><block> missing required attribute 'det'")
        det = _parse_det(det_str, context=f"<{context_tag}><block>@det")

        def decode_block_member(child: Element) -> BlockMember:
            if child.tag == "ref":
                ref_id = child.get("id")
                if ref_id is None:
                    raise ValueError(
                        f"<{context_tag}><block><ref> missing required attribute 'id'"
                    )

                try:
                    parts = ref_id.split("-")
                    if len(parts) != 2:
                        raise ValueError(
                            f"<{context_tag}><block><ref> attribute 'id' must be in format 'page-order'"
                        )
                    ref_page_index = int(parts[0])
                    ref_order = int(parts[1])
                except ValueError as e:
                    raise ValueError(
                        f"<{context_tag}><block><ref> attribute 'id' must contain valid integers"
                    ) from e

                if references_map is not None:
                    ref_key = (ref_page_index, ref_order)
                    if ref_key in references_map:
                        return references_map[ref_key]
                    else:
                        raise ValueError(
                            f"<{context_tag}><block><ref> references undefined reference: {ref_id}"
                        )
                else:
                    raise ValueError(
                        f"<{context_tag}><block><ref> cannot resolve reference without references_map"
                    )

            elif child.tag == "inline_expr":
                kind_attr = child.get("kind")
                if kind_attr is None:
                    raise ValueError(
                        f"<{context_tag}><block><inline_expr> missing required attribute 'kind'"
                    )
                kind = decode_expression_kind(kind_attr)
                expr_text = child.text if child.text is not None else ""
                return InlineExpression(kind=kind, content=expr_text)

            else:
                raise ValueError(
                    f"<{context_tag}><block> contains unknown element: <{child.tag}>"
                )

        blocks.append(
            BlockLayout(
                page_index=page_index,
                order=order,
                det=det,
                content=decode_content(block_el, decode_block_member),
            )
        )

    return blocks


def _encode_block_element(block: BlockLayout) -> Element:
    block_el = Element("block")
    block_el.set("page_index", str(block.page_index))
    block_el.set("order", str(block.order))
    block_el.set("det", ",".join(map(str, block.det)))
    encode_content(
        root=block_el,
        children=block.content,
        encode_payload=_encode_block_member,
    )
    return block_el


def _encode_block_member(part: BlockMember) -> Element:
    if isinstance(part, InlineExpression):
        expr_el = Element("inline_expr")
        expr_el.set("kind", encode_expression_kind(part.kind))
        expr_el.text = part.content
        return expr_el

    elif isinstance(part, Reference):
        ref_el = Element("ref")
        ref_el.set("id", f"{part.page_index}-{part.order}")
        return ref_el

    else:
        raise ValueError("Unknown BlockMember type")


def _encode_reference(ref: Reference) -> Element:
    ref_el = Element("ref")
    ref_el.set("id", f"{ref.page_index}-{ref.order}")

    mark_el = Element("mark")
    mark_el.text = str(ref.mark)
    ref_el.append(mark_el)

    for layout in ref.layouts:
        if isinstance(layout, AssetLayout):
            ref_el.append(_encode_asset(layout))
        else:
            ref_el.append(_encode_paragraph(layout))

    return ref_el


def _decode_reference(element: Element) -> Reference:
    ref_id = element.get("id")
    if ref_id is None:
        raise ValueError("<references><ref> missing required attribute 'id'")

    try:
        parts = ref_id.split("-")
        if len(parts) != 2:
            raise ValueError(
                "<references><ref> attribute 'id' must be in format 'page-order'"
            )
        page_index = int(parts[0])
        order = int(parts[1])
    except ValueError as e:
        raise ValueError(
            "<references><ref> attribute 'id' must contain valid integers"
        ) from e

    mark_el = element.find("mark")
    if mark_el is None or mark_el.text is None:
        raise ValueError("<references><ref> missing required <mark> element")
    mark_text = mark_el.text

    from .mark import transform2mark

    mark = transform2mark(mark_text)
    if mark is None:
        # 如果不是特殊标记，就使用原始字符串
        mark = mark_text

    layouts: list[AssetLayout | ParagraphLayout] = []
    for child in element:
        if child.tag == "mark":
            continue
        elif child.tag == "asset":
            layouts.append(_decode_asset(child))
        elif child.tag == "paragraph":
            layouts.append(_decode_paragraph(child, references_map=None))

    return Reference(
        page_index=page_index,
        order=order,
        mark=mark,
        layouts=layouts,
    )



from dataclasses import dataclass
from typing import Callable, Generic, Iterable, TypeVar

import ahocorasick

from ..language import is_latin_letter
from .text import normalize_text

_MAX_TOC_RATIO = 0.1
_TOC_HEAD_RATIO = 0.18
_TOC_SCORE_MIN_RATIO = 3.0  # 经验估计，以后再调整吧
_MIN_TOC_LIMIT = 3
_MIN_LATIN_TITLE_LENGTH = 6
_MIN_NON_LATIN_TITLE_LENGTH = 3


@dataclass
class PageRef:
    page_index: int
    score: float
    matched_titles: list["MatchedTitle"]


@dataclass
class MatchedTitle:
    text: str
    score: float
    references: list["TitleReference"]


@dataclass
class TitleReference:
    page_index: int
    order: int


# 使用统计学方式寻找文档中目录页所在页数范围。
# 目录页中的文本，会大规模与后续书页中的章节标题匹配，本函数使用此特征来锁定目录页。
def find_toc_pages(
    iter_titles: Callable[[], Iterable[list[tuple[int, str]]]],
    iter_page_bodies: Callable[[], Iterable[str]],
) -> list[PageRef]:
    matcher: _SubstringMatcher[tuple[int, int]] = (
        _SubstringMatcher()
    )  # (page_index, order)
    page_refs: list[PageRef] = []

    for page_index, titles_items in enumerate(iter_titles(), start=1):
        for order, title in titles_items:
            title = normalize_text(title)
            if _valid_title(title):
                matcher.register_substring(
                    substring=title,
                    payload=(page_index, order),
                )

    if matcher.substrings_count == 0:
        return []

    for page_index, body in enumerate(iter_page_bodies(), start=1):
        matched_titles: list[MatchedTitle] = []
        matched_substrings = matcher.match(normalize_text(body))

        # 每一个匹配的子串提供的分数为：该页匹配次数 / 该子串在文档中出现的总次数
        # 若匹配越多，当然说明此页更有可能是目录页。
        # 但若该子串在文档中大规模出现，例如书籍标题可能反复出现在页眉页脚，此时应该降低权重
        for substring, (matched_count, payloads) in matched_substrings.items():
            references: list[TitleReference] = [
                TitleReference(page_index=index, order=order)
                for index, order in payloads
                if index != page_index
            ]
            if references:
                matched_title = MatchedTitle(
                    text=substring,
                    score=matched_count / len(references),
                    references=references,
                )
                matched_titles.append(matched_title)

        page_refs.append(
            PageRef(
                page_index=page_index,
                matched_titles=matched_titles,
                score=sum(m.score for m in matched_titles),
            )
        )

    page_refs.sort(key=lambda x: x.score, reverse=True)
    max_diff = 0.0
    cut_position = 0

    for i in range(len(page_refs) - 1):
        diff = page_refs[i].score - page_refs[i + 1].score
        if diff > max_diff:
            max_diff = diff
            cut_position = i + 1

    toc_page_refs = page_refs[:cut_position]
    toc_page_refs.sort(key=lambda x: x.page_index)

    max_content_score = 0.0
    if cut_position < len(page_refs):
        max_content_score = page_refs[cut_position].score

    # DEBUG: 显示内容
    # for i, ref in enumerate(page_refs):
    #     if i == cut_position:
    #         print("\n----- TOC CANDIDATE CUT -----\n")
    #     print(f"[TOC PAGE] page_index={ref.page_index}, score={ref.score:.4f}")
    #     for title in ref.matched_titles:
    #         print(f"  [TITLE] score={title.score:.4f}, text={title.text}")
    #         for reference in title.references:
    #             print(f"    [REF] page_index={reference.page_index}, order={reference.order}")

    return _human_like_toc_filter(
        toc_page_refs=toc_page_refs,
        total_pages=len(page_refs),
        max_content_score=max_content_score,
    )


def _valid_title(title: str) -> bool:
    title = title.strip()
    if any(is_latin_letter(c) for c in title):
        return len(title) >= _MIN_LATIN_TITLE_LENGTH
    else:
        return len(title) >= _MIN_NON_LATIN_TITLE_LENGTH


def _human_like_toc_filter(
    toc_page_refs: list[PageRef],
    total_pages: int,
    max_content_score: float,
) -> list[PageRef]:
    max_toc_pages = max(_MIN_TOC_LIMIT, int(total_pages * _MAX_TOC_RATIO))
    max_toc_page_index = round(total_pages * _TOC_HEAD_RATIO)
    toc_page_refs = [
        ref for ref in toc_page_refs if ref.page_index <= max_toc_page_index
    ]
    if len(toc_page_refs) > max_toc_pages:
        toc_page_refs = toc_page_refs[:max_toc_pages]

    if not toc_page_refs:
        return toc_page_refs

    serial_refs: list[PageRef] = [toc_page_refs[0]]
    last_page_index = serial_refs[0].page_index

    for i in range(1, len(toc_page_refs)):
        ref = toc_page_refs[i]
        if ref.page_index == last_page_index + 1:
            serial_refs.append(ref)
            last_page_index = ref.page_index
        else:
            break

    if not serial_refs:
        return serial_refs

    serial_page_indexes = {ref.page_index for ref in serial_refs}
    for ref in toc_page_refs:
        if ref.page_index not in serial_page_indexes:
            max_content_score = max(max_content_score, ref.score)

    max_toc_score = serial_refs[0].score
    if max_toc_score < _TOC_SCORE_MIN_RATIO * max_content_score:
        return []  # 说明目录页不足以与非目录页拉开差距，不可贸然判断

    return serial_refs


_P = TypeVar("_P")


class _SubstringMatcher(Generic[_P]):
    def __init__(self):
        self._automaton = ahocorasick.Automaton()  # type: ignore[attr-defined]
        self._substrings_count: int = 0
        self._substring_to_payloads: dict[str, list[_P]] = {}
        self._finalized = False

    @property
    def substrings_count(self) -> int:
        return self._substrings_count

    def register_substring(self, substring: str, payload: _P) -> None:
        self._substrings_count += 1

        if substring not in self._substring_to_payloads:
            self._substring_to_payloads[substring] = []
            self._automaton.add_word(substring, substring)

        self._substring_to_payloads[substring].append(payload)
        self._finalized = False

    def match(self, text: str) -> dict[str, tuple[int, list[_P]]]:
        if not self._finalized:
            self._automaton.make_automaton()
            self._finalized = True

        match_counts: dict[str, int] = {}
        for _, substring in self._automaton.iter(text):
            match_counts[substring] = match_counts.get(substring, 0) + 1

        match_result: dict[str, tuple[int, list[_P]]] = {}
        for substring, count in match_counts.items():
            match_result[substring] = (count, self._substring_to_payloads[substring])

        return match_result



from .chapter import (
    AssetLayout,
    AssetRef,
    BlockLayout,
    BlockMember,
    Chapter,
    InlineExpression,
    ParagraphLayout,
    Reference,
    RefIdMap,
    decode,
    encode,
    references_to_map,
    search_references_in_chapter,
)
from .content import Content
from .generation import generate_chapter_files
from .mark import Mark, NumberClass, NumberStyle
from .reader import create_chapters_reader



def has_repetitive_ngrams(
    text: str,
    min_ngram: int,
    max_ngram: int,
    repeat_threshold: int,
) -> bool:
    """
    检测文本中是否存在大量重复的 n-gram 模式（字符级别）。

    这个方法用于检测 OCR 模型的 "Neural Text Degeneration" 问题。
    当 Transformer 模型训练不足或解码参数配置不当时，会生成重复的文本片段，
    例如 "这是一个例子这是一个例子这是一个例子..." 或
         "this is a test this is a test this is a test..."
    这种重复模式是无意义的噪声，需要被识别和过滤。

    采用字符级别分词以支持中文等没有空格分隔的语言。
    """
    if not text:
        return False

    chars = list(text)
    if len(chars) < min_ngram * repeat_threshold:
        return False

    for n in range(min_ngram, min(max_ngram + 1, len(chars) // repeat_threshold + 1)):
        for i in range(len(chars) - n * repeat_threshold + 1):
            ngram = tuple(chars[i : i + n])
            consecutive_count = 1
            pos = i + n
            while pos + n <= len(chars):
                next_ngram = tuple(chars[pos : pos + n])
                if next_ngram == ngram:
                    consecutive_count += 1
                    pos += n
                else:
                    break

            if consecutive_count >= repeat_threshold:
                return True

    return False



from dataclasses import dataclass
from typing import Generator
from xml.etree.ElementTree import Element

from ..common import indent


@dataclass
class TocInfo:
    content: list["Toc"]
    page_indexes: list[int]


@dataclass
class Toc:
    id: int
    page_index: int
    order: int
    level: int
    children: list["Toc"]


def iter_toc(toc_list: list[Toc]) -> Generator[Toc, None, None]:
    for toc in toc_list:
        yield toc
        yield from iter_toc(toc.children)


def encode(toc_info: TocInfo) -> Element:
    root = Element("toc")
    page_indexes_str = ",".join(str(idx) for idx in toc_info.page_indexes)
    root.set("page_indexes", page_indexes_str)

    def encode_item(toc: Toc, parent: Element) -> None:
        item = Element("item")
        item.set("id", str(toc.id))
        item.set("page_index", str(toc.page_index))
        item.set("order", str(toc.order))
        item.set("level", str(toc.level))

        for child in toc.children:
            encode_item(child, item)
        parent.append(item)

    for toc in toc_info.content:
        encode_item(toc, root)
    return indent(root)


def decode(element: Element) -> TocInfo:
    if element.tag != "toc":
        raise ValueError(f"Expected root tag 'toc', got '{element.tag}'")

    page_indexes_str = element.get("page_indexes")
    if page_indexes_str is None:
        raise ValueError("Missing 'page_indexes' attribute in toc")

    page_indexes = []
    if page_indexes_str:
        page_indexes = [int(idx) for idx in page_indexes_str.split(",")]

    def decode_item(item: Element) -> Toc:
        if item.tag != "item":
            raise ValueError(f"Expected tag 'item', got '{item.tag}'")

        id_str = item.get("id")
        page_index_str = item.get("page_index")
        order_str = item.get("order")
        level_str = item.get("level")

        if id_str is None:
            raise ValueError("Missing 'id' attribute in item")
        if page_index_str is None:
            raise ValueError("Missing 'page_index' attribute in item")
        if order_str is None:
            raise ValueError("Missing 'order' attribute in item")
        if level_str is None:
            raise ValueError("Missing 'level' attribute in item")

        children = [decode_item(child) for child in item]

        return Toc(
            id=int(id_str),
            page_index=int(page_index_str),
            order=int(order_str),
            level=int(level_str),
            children=children,
        )

    return TocInfo(
        content=[decode_item(item) for item in element],
        page_indexes=page_indexes,
    )



from pathlib import Path
from shutil import copy
from typing import Generator

from ...metering import AbortedCheck, check_aborted
from ...sequence import (
    Reference,
    create_chapters_reader,
    references_to_map,
    search_references_in_chapter,
)
from .layouts import render_layouts


def render_markdown_file(
    chapters_path: Path,
    assets_path: Path,
    output_path: Path,
    output_assets_path: Path,
    cover_path: Path | None,
    aborted: AbortedCheck,
):
    assets_ref_path = output_assets_path
    if not assets_ref_path.is_absolute():
        output_assets_path = output_path.parent / output_assets_path

    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_assets_path.mkdir(parents=True, exist_ok=True)
    read_chapters = create_chapters_reader(chapters_path)

    references: list[Reference] = []
    for chapter in read_chapters():
        references.extend(search_references_in_chapter(chapter))

    references.sort(key=lambda ref: (ref.page_index, ref.order))
    ref_id_to_number = references_to_map(references)

    with open(output_path, "w", encoding="utf-8") as f:
        need_blank_line = False
        for chapter in read_chapters():
            check_aborted(aborted)

            if need_blank_line:
                need_blank_line = False
                f.write("\n\n")

            for part in render_layouts(
                layouts=chapter.layouts,
                assets_path=assets_path,
                output_assets_path=output_assets_path,
                asset_ref_path=assets_ref_path,
                toc_level=chapter.level,
                ref_id_to_number=ref_id_to_number,
            ):
                f.write(part)
                need_blank_line = True

        check_aborted(aborted)
        for part in _render_footnotes_section(
            references=references,
            assets_path=assets_path,
            output_assets_path=output_assets_path,
            asset_ref_path=assets_ref_path,
        ):
            f.write(part)

    if cover_path is not None:
        copy(
            src=cover_path,
            dst=output_assets_path / cover_path.name,
        )


def _render_footnotes_section(
    references: list[Reference],
    assets_path: Path,
    output_assets_path: Path,
    asset_ref_path: Path,
) -> Generator[str, None, None]:
    if not references:
        return
    yield "\n\n---\n\n## References"
    for i, ref in enumerate(references, 1):
        yield "\n\n"
        yield f"[^{i}]:  "
        yield from render_layouts(
            layouts=ref.layouts,
            assets_path=assets_path,
            output_assets_path=output_assets_path,
            toc_level=0,
            asset_ref_path=asset_ref_path,
        )



from typing import Iterable


def avg(values: Iterable[float]) -> float:
    total = 0.0
    count = 0
    for value in values:
        total += value
        count += 1

    if count == 0:
        raise ValueError("Cannot calculate average of empty sequence")

    return total / count


def median(values: Iterable[float]) -> float:
    values = list(values)
    if not values:
        raise ValueError("Cannot calculate median of empty sequence")

    sorted_values = sorted(values)
    n = len(sorted_values)

    if n % 2 == 1:
        return sorted_values[n // 2]
    else:
        mid1 = sorted_values[n // 2 - 1]
        mid2 = sorted_values[n // 2]
        return (mid1 + mid2) / 2



from datetime import datetime, timezone
from os import PathLike
from pathlib import Path
from typing import Protocol, cast, runtime_checkable

from PIL import Image

from ..error import PDFError
from .types import PDFDocumentMetadata


@runtime_checkable
class PDFDocument(Protocol):
    @property
    def pages_count(self) -> int: ...

    def metadata(self) -> PDFDocumentMetadata: ...

    def page_size(self, page_index: int) -> tuple[float, float]: ...

    def render_page(self, page_index: int, dpi: int) -> Image.Image: ...

    def close(self) -> None: ...


@runtime_checkable
class PDFHandler(Protocol):
    def open(self, pdf_path: Path) -> PDFDocument: ...


class DefaultPDFHandler:
    def __init__(self, poppler_path: PathLike | str | None = None) -> None:
        self._poppler_path: Path | None = None
        if poppler_path is not None:
            self._poppler_path = Path(poppler_path)

    def open(self, pdf_path: Path) -> PDFDocument:
        return DefaultPDFDocument(
            pdf_path=pdf_path,
            poppler_path=self._poppler_path,
        )


_POINTS_PER_INCH = 72.0


class DefaultPDFDocument:
    def __init__(self, pdf_path: Path, poppler_path: Path | None) -> None:
        import pypdf

        self._pdf_path = pdf_path
        self._poppler_path: Path | None = poppler_path
        self._reader = pypdf.PdfReader(str(pdf_path))
        self._pages_count: int | None = None

    @property
    def pages_count(self) -> int:
        if self._pages_count is None:
            self._pages_count = len(self._reader.pages)
        return self._pages_count

    def metadata(self) -> PDFDocumentMetadata:
        try:
            metadata = self._reader.metadata
            title = (
                str(metadata.get("/Title"))
                if metadata and metadata.get("/Title")
                else None
            )
            description = (
                str(metadata.get("/Subject"))
                if metadata and metadata.get("/Subject")
                else None
            )
            authors: list[str] = []

            if metadata and "/Author" in metadata:
                author_obj = metadata["/Author"]
                if author_obj:
                    author_str = str(author_obj)
                    for sep in (";", ",", "&"):
                        if sep in author_str:
                            authors = [
                                a.strip() for a in author_str.split(sep) if a.strip()
                            ]
                            break
                    if not authors:
                        authors = [author_str.strip()]

            modified = datetime.now(timezone.utc)
            if metadata and "/ModDate" in metadata:
                try:
                    mod_date_obj = metadata["/ModDate"]
                    if mod_date_obj:
                        mod_date_str = str(mod_date_obj)
                        # PDF 日期格式: D:YYYYMMDDHHmmSSOHH'mm'
                        if mod_date_str.startswith("D:"):
                            mod_date_str = mod_date_str[2:]
                        # 简化处理：只取年月日时分秒
                        if len(mod_date_str) >= 14:
                            year = int(mod_date_str[0:4])
                            month = int(mod_date_str[4:6])
                            day = int(mod_date_str[6:8])
                            hour = int(mod_date_str[8:10])
                            minute = int(mod_date_str[10:12])
                            second = int(mod_date_str[12:14])
                            modified = datetime(
                                year,
                                month,
                                day,
                                hour,
                                minute,
                                second,
                                tzinfo=timezone.utc,
                            )
                except (ValueError, IndexError):
                    pass  # 解析失败则使用当前时间

            return PDFDocumentMetadata(
                title=title,
                description=description,
                publisher=None,  # PDF 标准元数据中通常没有 publisher
                isbn=None,  # PDF 标准元数据中通常没有 ISBN
                authors=authors,
                editors=[],  # PDF 标准元数据中通常没有 editors
                translators=[],  # PDF 标准元数据中通常没有 translators
                modified=modified,
            )
        except Exception as error:
            raise PDFError(
                "Failed to extract PDF metadata.", page_index=None
            ) from error

    def page_size(self, page_index: int) -> tuple[float, float]:
        try:
            page = self._reader.pages[page_index - 1]
            width_inch = float(page.mediabox.width) / _POINTS_PER_INCH
            height_inch = float(page.mediabox.height) / _POINTS_PER_INCH
            return (width_inch, height_inch)
        except Exception as error:
            raise PDFError(
                f"Failed to get page size for page {page_index}.", page_index=page_index
            ) from error

    def render_page(self, page_index: int, dpi: int) -> Image.Image:
        from pdf2image import convert_from_path
        from pdf2image.exceptions import PDFInfoNotInstalledError

        poppler_path: str | None
        if self._poppler_path:
            poppler_path = str(self._poppler_path)
        else:
            poppler_path = None  # use poppler in system PATH

        try:
            images: list[Image.Image] = convert_from_path(
                str(self._pdf_path),
                dpi=dpi,
                first_page=page_index,
                last_page=page_index,
                poppler_path=cast(str, poppler_path),
            )
        except PDFInfoNotInstalledError as error:
            if self._poppler_path:
                error_message = (
                    f"Poppler not found at specified path: {self._poppler_path}"
                )
            else:
                error_message = "Poppler not found in PATH. Either not installed or PATH is not configured correctly."
            raise PDFError(error_message, page_index) from error

        if not images:
            raise RuntimeError(f"Failed to render page {page_index}")

        image = images[0]
        if image.mode != "RGB":
            image = image.convert("RGB")

        return image

    def close(self) -> None:
        self._reader.stream.close()



from typing import Callable, Generator, Iterable

from ...language import is_chinese_char
from .types import HTMLTag, P


def render_markdown_paragraph(
    children: list[str | P | HTMLTag[P]],
    render_payload: Callable[[P | str], Iterable[str]],
) -> Generator[str, None, None]:
    yield from _normalize_paragraph(
        parts=_render_markdown_children(
            children=children,
            render_payload=render_payload,
        ),
    )


def _render_markdown_children(
    children: list[str | P | HTMLTag[P]],
    render_payload: Callable[[P | str], Iterable[str]],
) -> Generator[str, None, None]:
    for child in children:
        if isinstance(child, HTMLTag):
            yield from _render_html_tag(child, render_payload)
        else:
            yield from render_payload(child)


def _render_html_tag(
    tag: HTMLTag[P], render_payload: Callable[[P | str], Iterable[str]]
) -> Generator[str, None, None]:
    tag_name = tag.definition.name

    if not tag.children:
        yield "<"
        yield tag_name
        yield from _render_attributes(tag.attributes)
        yield " />"
    else:
        yield "<"
        yield tag_name
        yield from _render_attributes(tag.attributes)
        yield ">"

        yield from render_markdown_paragraph(tag.children, render_payload)

        yield "</"
        yield tag_name
        yield ">"


def _render_attributes(attributes: list[tuple[str, str]]) -> Generator[str, None, None]:
    for name, value in attributes:
        yield " "
        yield name
        if value:
            yield '="'
            yield from _escape_attribute(value)
            yield '"'


def _escape_attribute(value: str) -> Generator[str, None, None]:
    for char in value:
        if char == "&":
            yield "&amp;"
        elif char == '"':
            yield "&quot;"
        elif char == "<":
            yield "&lt;"
        elif char == ">":
            yield "&gt;"
        else:
            yield char


def _normalize_paragraph(parts: Iterable[str]) -> Generator[str, None, None]:
    last_char: str | None = None
    is_line_head = True
    for part in _split_enters(parts):
        if part != "\n":
            if is_line_head:
                is_line_head = False
                part = part.lstrip()
                if (
                    part
                    and last_char is not None
                    and (not is_chinese_char(last_char) or not is_chinese_char(part[0]))
                ):
                    yield " "
            if part:
                yield part
                part = part.rstrip()
                if part:
                    last_char = part[-1]
        else:
            is_line_head = True


def _split_enters(parts: Iterable[str]) -> Generator[str, None, None]:
    for part in parts:
        if not part:
            continue
        split_parts = part.splitlines()
        yield split_parts[0]
        for i in range(1, len(split_parts)):
            yield "\n"
            yield split_parts[i]



import re
from dataclasses import dataclass
from typing import Generator, Iterable, cast

from ..common import ASSET_TAGS, AssetRef
from ..expression import ExpressionKind, ParsedItem, parse_latex_expressions
from ..language import is_latin_letter
from ..markdown.paragraph import parse_raw_markdown
from ..pdf import TITLE_TAGS, PageLayout
from .chapter import AssetLayout, BlockLayout, InlineExpression, ParagraphLayout
from .content import Content, expand_text_in_content, first, last
from .reading_serials import split_reading_serials

_ASSET_CAPTION_TAGS = tuple(f"{t}_caption" for t in ASSET_TAGS)

# to see https://github.com/opendatalab/MinerU/blob/fa1149cd4abf9db5e0f13e4e074cdb568be189f4/mineru/utils/span_pre_proc.py#L247
_LINE_STOP_FLAGS = (
    ".",
    "!",
    "?",
    "。",
    "！",
    "？",
    ")",
    "）",
    """, """,
    ";",
    "；",
    "]",
    "】",
    "}",
    ">",
    "》",
)

_LINE_CONTINUE_FLAGS = (
    "[",
    "【",
    "{",
    "<",
    "《",
    "、",
    ",",
    "，",
)

_LINK_FLAGS = (
    "‐",
    "‑",
    "‒",
    "–",
    "—",
    "―",
)

_MARKDOWN_HEAD_PATTERN = re.compile(r"^#+\s+")
_TABLE_PATTERN = re.compile(r"<table[^>]*>.*?</table>", re.IGNORECASE | re.DOTALL)


@dataclass
class _LastTail:
    page_para: ParagraphLayout
    override: list[AssetLayout]


@dataclass
class _AssetHolder:
    page_index: int
    ref: AssetRef
    det: tuple[int, int, int, int]
    title: str | None
    content: str
    caption: str | None
    hash: str | None


class Jointer:
    def __init__(self, layouts: Iterable[tuple[int, list[PageLayout]]]) -> None:
        self._layouts = layouts

    def execute(self) -> Generator[ParagraphLayout | AssetLayout, None, None]:
        last_tail: _LastTail | None = None

        for page_index, raw_layouts in self._iter_layout_serials():
            # 此处为完成如下业务要求：
            # 1. 当阅读序列跨越 group（跨页、跨分栏、跨因图片而挤变形拆分的段落）时，必须对连接处验证。若它们是被拆分的自然段，则拼起来。
            # 2. 因为插图、表格而拆分的自然段，需将插图存起来接到完整的自然段最后，而不是任其分割自然段。
            layouts = list(self._join_and_handle_asset_layouts(page_index, raw_layouts))
            head, body, tail = self._split_layouts(layouts)

            if not body:
                if last_tail:
                    last_tail.override.extend(head)
                    last_tail.override.extend(tail)
                else:
                    yield from head
                    yield from tail
                continue

            first_layout = cast(ParagraphLayout, body[0])
            if last_tail and self._can_merge_paragraphs(
                last_tail.page_para, first_layout
            ):
                last_tail.page_para.blocks.extend(first_layout.blocks)
                del body[0]

            if not body:
                if last_tail:
                    last_tail.override.extend(head)
                    last_tail.override.extend(tail)
                else:
                    yield from head
                    yield from tail
                continue

            # 至此，连续吞并段落的流程遇阻而结束
            if last_tail:
                _normalize_paragraph_content(last_tail.page_para)
                yield last_tail.page_para
                yield from last_tail.override
                last_tail = None

            yield from head
            for i in range(len(body) - 1):
                yield body[i]

            last_tail = _LastTail(
                page_para=cast(ParagraphLayout, body[-1]),
                override=list(tail),
            )

        if last_tail:
            _normalize_paragraph_content(last_tail.page_para)
            yield last_tail.page_para
            yield from last_tail.override

    def _iter_layout_serials(
        self,
    ) -> Generator[tuple[int, list[PageLayout]], None, None]:
        for page_index, raw_layouts in self._layouts:
            for layouts in split_reading_serials(raw_layouts):
                yield page_index, layouts

    def _split_layouts(self, layouts: list[ParagraphLayout | AssetLayout]):
        head: list[AssetLayout] = []
        tail: list[AssetLayout] = []

        for layout in layouts:
            if isinstance(layout, ParagraphLayout):
                break
            head.append(layout)

        for i in range(len(layouts) - 1, -1, -1):
            if i < len(head):
                break
            layout = layouts[i]
            if isinstance(layout, ParagraphLayout):
                break
            tail.append(layout)

        tail.reverse()
        body = layouts[len(head) : len(layouts) - len(tail)]

        return head, body, tail

    def _join_and_handle_asset_layouts(
        self, page_index, layouts: list[PageLayout]
    ) -> Generator[ParagraphLayout | AssetLayout, None, None]:
        # layout 可能被后续处理，必须等待所有 layout 处理完毕
        for layout in list(
            self._join_asset_layouts(
                page_index=page_index,
                layouts=layouts,
            )
        ):
            if not isinstance(layout, _AssetHolder):
                yield layout
                continue

            if layout.ref == "equation":
                _normalize_equation(layout)
            if layout.ref == "table":
                _normalize_table(layout)

            yield AssetLayout(
                page_index=page_index,
                ref=layout.ref,
                det=layout.det,
                title=_parse_block_content(layout.title),
                content=_parse_block_content(layout.content),
                caption=_parse_block_content(layout.caption),
                hash=layout.hash,
            )

    def _join_asset_layouts(self, page_index, layouts: list[PageLayout]):
        last_asset: _AssetHolder | None = None
        for layout in layouts:
            if layout.ref in ASSET_TAGS:
                if last_asset:
                    yield last_asset
                last_asset = _AssetHolder(
                    page_index=page_index,
                    ref=layout.ref,
                    det=layout.det,
                    title=None,
                    content=layout.text,
                    caption=None,
                    hash=layout.hash,
                )
            elif layout.ref in _ASSET_CAPTION_TAGS:
                if last_asset:
                    if last_asset.caption:
                        last_asset.caption += "\n" + layout.text
                    else:
                        last_asset.caption = layout.text
            else:
                if last_asset:
                    yield last_asset
                    last_asset = None
                if layout.ref in TITLE_TAGS:
                    # 将 Markdown 标题前的 `##` 之类的符号删除，DeepSeek OCR 总会生成这种符号
                    layout.text = _MARKDOWN_HEAD_PATTERN.sub("", layout.text)

                yield ParagraphLayout(
                    ref=layout.ref,
                    level=-1,
                    blocks=[
                        BlockLayout(
                            page_index=page_index,
                            order=layout.order,
                            det=layout.det,
                            content=_parse_block_content(layout.text),
                        )
                    ],
                )
        if last_asset:
            yield last_asset

    # too see https://github.com/opendatalab/MinerU/blob/fa1149cd4abf9db5e0f13e4e074cdb568be189f4/mineru/backend/pipeline/para_split.py#L253
    def _can_merge_paragraphs(
        self, para1: ParagraphLayout, para2: ParagraphLayout
    ) -> bool:
        if para1.ref != "text":
            return False
        if para1.ref != para2.ref:
            return False

        block1 = para1.blocks[-1]
        block2 = para2.blocks[0]

        text1 = last(block1.content)
        text2 = first(block2.content)
        if not isinstance(text1, str) or not isinstance(text2, str):
            return False

        text1_stripped = text1.rstrip()
        text2_stripped = text2.lstrip()
        if not text1_stripped or not text2_stripped:
            return False

        # 条件1：前一个段落如果以句尾符号结尾，说明是完整段落，不应合并
        if text1_stripped.endswith(_LINE_STOP_FLAGS):
            return False

        # 条件2：前一个段落结束的符号明显表明句子未结束，则必须合并
        if text1_stripped.endswith(_LINE_CONTINUE_FLAGS):
            return True

        first_char = text2_stripped[0]

        # 条件3：下一个段落的第一个字符不是数字
        # 如果以数字开头，可能是编号列表的新段落（如"1. xxx"）
        if first_char.isdigit():
            return False

        # 条件4：下一个段落的第一个字符不是大写字母
        # 如果以大写字母开头，可能是新段落的开始（特别是英文）
        if first_char.isupper():
            return False

        # 条件5：如果 para1 结尾是拉丁字母 + `-`，para2 开头是拉丁字母，则允许合并（跨段单词拼接）
        if is_latin_letter(text2[0]):
            if (
                len(text1) >= 2
                and text1[-1] in _LINK_FLAGS
                and is_latin_letter(text1[-2])
            ):
                return True
            if is_latin_letter(text1[-1]):
                return False

        return True


def _normalize_equation(layout: _AssetHolder):
    if layout.ref != "equation" or not layout.content:
        return

    found_first_expression: bool = False
    expression_content: str = ""
    prefix_texts: list[str] = []
    tail_items: list[ParsedItem] = []

    for item in parse_latex_expressions(layout.content):
        if not found_first_expression and item.kind != ExpressionKind.TEXT:
            expression_content = item.content
            found_first_expression = True
        elif found_first_expression:
            tail_items.append(item)
        else:
            prefix_texts.append(item.content)

    if not found_first_expression:
        return

    if layout.title is not None:
        prefix_texts.insert(0, layout.title)

    if layout.caption is not None:
        tail_items.append(ParsedItem(kind=ExpressionKind.TEXT, content=layout.caption))

    if prefix_texts:
        layout.title = "".join(prefix_texts)

    layout.content = expression_content

    if tail_items:
        layout.caption = "".join(item.reverse() for item in tail_items)


def _normalize_table(layout: _AssetHolder):
    found_table_content: str | None = None
    head_buffer: list[str] = []
    tail_buffer: list[str] = []

    for part in (layout.title, "\n", layout.content, "\n", layout.caption):
        if not part:
            continue

        table_match = _TABLE_PATTERN.search(part)
        if not table_match:
            if found_table_content is None:
                head_buffer.append(part)
            else:
                tail_buffer.append(part)
            continue

        table_start = table_match.start()
        table_end = table_match.end()

        table_content = part[table_start:table_end]
        before = part[:table_start].rstrip()
        after = part[table_end:].lstrip()

        if before.strip():
            head_buffer.append(before)
        if after.strip():
            tail_buffer.append(after)

        found_table_content = table_content

    if not found_table_content:
        return

    head = "".join(head_buffer).strip()
    tail = "".join(tail_buffer).strip()

    layout.title = head if head else None
    layout.caption = tail if tail else None
    layout.content = found_table_content


# 将单词的连接符 `-` 删去，并将后半节单词移到前面一段拼接
def _normalize_paragraph_content(paragraph: ParagraphLayout):
    if len(paragraph.blocks) < 2:
        return

    for i in range(1, len(paragraph.blocks)):
        block1 = paragraph.blocks[i - 1]
        block2 = paragraph.blocks[i]

        text1 = last(block1.content)
        text2 = first(block2.content)
        if not isinstance(text1, str) or not isinstance(text2, str):
            continue

        text1 = text1.rstrip()
        text2 = text2.lstrip()
        if not _is_splitted_word(text1, text2):
            continue

        tail_end = 0
        for j in range(len(text2)):
            if is_latin_letter(text2[j]):
                tail_end = j + 1
            else:
                break

        block1.content[-1] = text1[:-1] + text2[:tail_end]
        block2.content[0] = text2[tail_end:].lstrip()
        if not block2.content[0]:
            del block2.content[0]

    # 极端情况下 block2 会因为单词被移走而被清空。此时要将其整个删去。
    paragraph.blocks = [block for block in paragraph.blocks if block.content]


def _parse_block_content(text: str | None) -> Content:
    if not text:
        return []

    root_content: Content = parse_raw_markdown(text)

    def expand_text(text: str):
        for item in parse_latex_expressions(text):
            if item.kind != ExpressionKind.TEXT:
                yield InlineExpression(
                    kind=item.kind,
                    content=item.content,
                )
            elif item.content:  # Only add non-empty strings
                yield item.content

    expand_text_in_content(
        content=root_content,
        expand=expand_text,
    )
    return root_content


def _is_splitted_word(text1: str, text2: str) -> bool:
    return (
        len(text1) >= 2
        and text1[-1] in _LINK_FLAGS
        and is_latin_letter(text1[-2])
        and is_latin_letter(text2[0])
    )



from pathlib import Path
from typing import Callable, Generator, Iterable

from ..common import XMLReader, read_xml
from .chapter import Chapter, decode


def create_chapters_reader(chapters_path: Path) -> Callable[[], Iterable[Chapter]]:
    chapters: XMLReader[Chapter] = XMLReader(
        prefix="chapter",
        dir_path=chapters_path,
        decode=decode,
    )
    head_path = chapters_path / "chapter_head.xml"

    def generate() -> Generator[Chapter, None, None]:
        if head_path.exists():
            root = read_xml(head_path)
            try:
                yield decode(root)
            except Exception as e:
                raise ValueError(f"Failed to decode from: {head_path}") from e

        yield from chapters.read()

    return generate



from .parser import parse_raw_markdown
from .render import render_markdown_paragraph
from .tags import (
    HTMLTagDefinition,
    is_protocol_allowed,
    is_tag_filtered,
    is_tag_ignored,
    tag_definition,
)
from .types import HTMLTag, P, decode, encode, flatten



from epub_generator import BookMeta, LaTeXRender, TableRender

from .error import (
    IgnoreOCRErrorsChecker,
    IgnorePDFErrorsChecker,
    InterruptedError,
    OCRError,
    PDFError,
)
from .functions import predownload_models, transform_epub, transform_markdown
from .metering import AbortedCheck, InterruptedKind
from .pdf import (
    DeepSeekOCRSize,
    DefaultPDFDocument,
    DefaultPDFHandler,
    OCREvent,
    OCREventKind,
    PDFDocument,
    PDFDocumentMetadata,
    PDFHandler,
    pdf_pages_count,
)
from .transform import OCRTokensMetering, Transform



from dataclasses import dataclass
from pathlib import Path

from ..common import XMLReader, avg, read_xml, split_by_cv
from ..config import MAX_TITLE_CV
from ..pdf import TITLE_TAGS, Page, PageLayout
from ..pdf import decode as decode_as_page
from .text import normalize_text
from .toc_pages import PageRef

_MAX_LEVELS = 4
_MAX_TOC_CV = 0.75  # 不宜过小导致过多分组

Ref2Level = dict[tuple[int, int], int]  # key: (page_index, order) value: level


def analyse_title_levels(pages: XMLReader[Page]) -> Ref2Level:
    return _extract_content_title_levels(pages)


def analyse_toc_levels(
    pages: XMLReader[Page], pages_path: Path, toc_pages: list[PageRef]
) -> Ref2Level:
    ref2meta, toc_page_indexes = _extract_ref2meta(
        pages_path=pages_path,
        toc_pages=toc_pages,
    )
    ref2global_level = _extract_content_title_levels(
        pages=pages,
        disable_page_indexes=toc_page_indexes,
        ref2meta=ref2meta,
    )
    toc_level_offset = _extract_toc_level_offset(
        ref2meta=ref2meta,
        ref2level=ref2global_level,
    )
    ref2level: Ref2Level = {}
    for (page_index, order), meta in ref2meta.items():
        level_offset = toc_level_offset.get(meta.toc_page_index, None)
        if (
            level_offset is not None
        ):  # toc_level_offset 比 ref2meta 范围小，目录页子圈会被排除
            global_level = meta.relative_level + level_offset
            ref2level[(page_index, order)] = global_level

    return ref2level


@dataclass
class _Hook:
    layout: PageLayout
    references: list[tuple[int, int]]  # (page_index, order)


@dataclass
class _TitleMeta:
    toc_page_index: int
    relative_level: int
    collected_global_levels: list[int]


_Ref2Meta = dict[tuple[int, int], _TitleMeta]


def _extract_ref2meta(
    pages_path: Path, toc_pages: list[PageRef]
) -> tuple[_Ref2Meta, set[int]]:
    ref2meta: _Ref2Meta = {}  # key: (page_index, order)
    toc_page_indexes: set[int] = set()

    for ref in toc_pages:
        toc_page_indexes.add(ref.page_index)
        grouped_hooks = _analyse_toc_page_hooks(
            ref=ref,
            page_path=Path(pages_path / f"page_{ref.page_index}.xml"),
        )
        for level, hooks in enumerate(grouped_hooks):
            for hook in sorted(hooks, key=lambda x: x.layout.order):
                for page_index, order in hook.references:
                    if (page_index, order) not in ref2meta:
                        ref2meta[(page_index, order)] = _TitleMeta(
                            toc_page_index=ref.page_index,
                            relative_level=level,
                            collected_global_levels=[],
                        )
    return ref2meta, toc_page_indexes


def _analyse_toc_page_hooks(ref: PageRef, page_path: Path) -> list[list[_Hook]]:
    page = decode_as_page(read_xml(page_path))
    hooks_items: list[tuple[float, _Hook]] = []

    for layout in page.body_layouts:
        layout_text = normalize_text(layout.text)
        references_set: set[tuple[int, int]] = set()
        for title in ref.matched_titles:
            if title.text not in layout_text:
                continue
            for reference in title.references:
                references_set.add(
                    (
                        reference.page_index,
                        reference.order,
                    )
                )
        if not references_set:
            continue
        _, top, _, bottom = layout.det
        height = bottom - top
        hooks_items.append(
            (
                height,
                _Hook(
                    layout=layout,
                    references=list(references_set),
                ),
            )
        )
    hooks = split_by_cv(
        payload_items=hooks_items,
        max_groups=_MAX_LEVELS,
        max_cv=_MAX_TOC_CV,
    )
    hooks.reverse()  # 字体最大的是 Level 0，故颠倒
    return hooks


def _extract_content_title_levels(
    pages: XMLReader[Page],
    disable_page_indexes: set[int] | None = None,
    ref2meta: _Ref2Meta | None = None,
) -> Ref2Level:
    title_items: list[tuple[float, tuple[int, int]]] = []
    for page in pages.read():
        if disable_page_indexes and page.index in disable_page_indexes:
            continue
        for layout in page.body_layouts:
            if layout.ref not in TITLE_TAGS:
                continue
            ref = (page.index, layout.order)
            if ref2meta and ref not in ref2meta:
                continue
            _, top, _, bottom = layout.det
            height = bottom - top
            title_items.append((height, ref))

    title2levels: Ref2Level = {}
    for level, refs in enumerate(
        reversed(
            split_by_cv(  # 字体最大的是 Level 0，故颠倒
                payload_items=title_items,
                max_groups=_MAX_LEVELS,
                max_cv=MAX_TITLE_CV,
            )
        )
    ):
        for ref in refs:
            title2levels[ref] = level

    return title2levels


def _extract_toc_level_offset(
    ref2meta: _Ref2Meta, ref2level: Ref2Level
) -> dict[int, int]:
    # 目录可能有多页，每页自己的 Level 只能表示相对关系，这里计算出全局目录的 level
    for page_index, order in sorted(ref2level.keys()):
        meta = ref2meta[(page_index, order)]
        level = ref2level[(page_index, order)]
        meta.collected_global_levels.append(level)

    page2metas: dict[int, list[_TitleMeta]] = {}
    for meta in ref2meta.values():
        metas = page2metas.get(meta.toc_page_index, None)
        if metas is None:
            metas = []
            page2metas[meta.toc_page_index] = metas
        metas.append(meta)

    avg_level_items: list[tuple[float, int]] = []  # (avg_level, page_index)
    for page_index, metas in page2metas.items():
        metas.sort(key=lambda x: x.relative_level)
        meta = metas[0]
        levels = metas[0].collected_global_levels
        if levels:
            avg_level = avg(levels)
            avg_level_items.append((avg_level, page_index))

    toc_level_offset: dict[int, int] = {}
    for offset, page_indexes in enumerate(
        split_by_cv(
            payload_items=avg_level_items,
            max_groups=_MAX_LEVELS,
            max_cv=_MAX_TOC_CV,
        )
    ):
        # 大多数情况下只有一组，便只有 offset=0 的情况。
        for page_index in page_indexes:
            toc_level_offset[page_index] = offset

    return toc_level_offset



from typing import Callable, Generator, Iterable

from ..markdown.paragraph import HTMLTag
from .chapter import BlockMember

Content = list[str | BlockMember | HTMLTag[BlockMember]]


def first(content: Content) -> None | str | BlockMember:
    if not content:
        return None
    element = content[0]
    if isinstance(element, HTMLTag):
        return first(element.children)
    else:
        return element


def last(content: Content) -> None | str | BlockMember:
    if not content:
        return None
    element = content[-1]
    if isinstance(element, HTMLTag):
        return last(element.children)
    else:
        return element


def join_texts_in_content(content: Content) -> None:
    for sub_content in _search_content(content):
        i: int = 0
        while i < len(sub_content) - 1:
            part1 = sub_content[i]
            part2 = sub_content[i + 1]
            if isinstance(part1, str) and isinstance(part2, str):
                sub_content[i] = part1 + part2
                del sub_content[i + 1]
            else:
                i += 1


def expand_text_in_content(
    content: Content,
    expand: Callable[[str], Iterable[str | BlockMember]],
) -> None:
    for sub_content in _search_content(content):
        i: int = 0
        while i < len(sub_content):
            part = sub_content[i]
            if isinstance(part, str):
                del sub_content[i]
                for expanded_part in expand(part):
                    sub_content.insert(i, expanded_part)
                    i += 1
            else:
                i += 1


def _search_content(content: Content) -> Generator[Content, None, None]:
    for child in content:
        if isinstance(child, HTMLTag):
            yield from _search_content(child.children)
    yield content



from pathlib import Path
from typing import Generator, Literal

from epub_generator import (
    BookMeta,
    ChapterGetter,
    EpubData,
    Footnote,
    Formula,
    Image,
    LaTeXRender,
    Mark,
    Table,
    TableRender,
    TextBlock,
    TextKind,
    generate_epub,
)
from epub_generator import (
    Chapter as ChapterRecord,
)
from epub_generator import (
    HTMLTag as EpubHTMLTag,
)

from ..markdown.paragraph import HTMLTag, flatten
from ..metering import AbortedCheck, check_aborted
from ..pdf import TITLE_TAGS
from ..sequence import (
    AssetLayout,
    Chapter,
    InlineExpression,
    ParagraphLayout,
    Reference,
    create_chapters_reader,
    references_to_map,
    search_references_in_chapter,
)
from .latex_to_text import latex_to_plain_text
from .toc_collection import TocCollection


def render_epub_file(
    chapters_path: Path,
    toc_path: Path | None,
    assets_path: Path,
    epub_path: Path,
    cover_path: Path | None,
    book_meta: BookMeta | None,
    lan: Literal["zh", "en"],
    table_render: TableRender,
    latex_render: LaTeXRender,
    inline_latex: bool,
    aborted: AbortedCheck,
):
    read_chapters = create_chapters_reader(chapters_path)
    references: list[Reference] = []
    for chapter in read_chapters():
        references.extend(search_references_in_chapter(chapter))

    references.sort(key=lambda ref: (ref.page_index, ref.order))
    ref_id_to_number = references_to_map(references)
    get_head: ChapterGetter | None = None
    toc_collection = TocCollection(toc_path)

    for chapter in read_chapters():

        def get_chapter(ch=chapter):
            return _convert_chapter_to_epub(
                chapter=ch,
                assets_path=assets_path,
                inline_latex=inline_latex,
                ref_id_to_number=ref_id_to_number,
            )

        if chapter.id is None:
            get_head = get_chapter
        elif chapter.layouts:
            first_layout = chapter.layouts[0]
            if (
                isinstance(first_layout, ParagraphLayout)
                and first_layout.ref in TITLE_TAGS
            ):
                title = "".join(_iter_text_in_title(first_layout)).strip()
                if not title:
                    title = "Untitled"
                have_body = len(chapter.layouts) > 1
                toc_collection.collect(
                    toc_id=chapter.id,
                    title=title,
                    have_body=have_body,
                    get_chapter=get_chapter if have_body else None,
                )

    epub_data = EpubData(
        meta=book_meta,
        get_head=get_head,
        chapters=toc_collection.normalize().target,
        cover_image_path=cover_path,
    )
    check_aborted(aborted)
    generate_epub(
        epub_data=epub_data,
        epub_file_path=epub_path,
        lan=lan,
        table_render=table_render,
        latex_render=latex_render,
        assert_not_aborted=lambda: check_aborted(aborted),
    )


def _iter_text_in_title(title_layout: ParagraphLayout):
    for block in title_layout.blocks:
        for item in flatten(block.content):
            if isinstance(item, str):
                yield item


def _convert_chapter_to_epub(
    chapter: Chapter,
    assets_path: Path,
    inline_latex: bool,
    ref_id_to_number: dict,
) -> ChapterRecord:
    elements = []
    footnotes = []

    for layout in chapter.layouts:
        if isinstance(layout, AssetLayout):
            asset_element = _convert_asset_to_epub(
                asset=layout,
                assets_path=assets_path,
                inline_latex=inline_latex,
                ref_id_to_number=ref_id_to_number,
            )
            if asset_element:
                elements.append(asset_element)
        elif isinstance(layout, ParagraphLayout):
            content: list[str | Formula | Mark | EpubHTMLTag] = []
            for block in layout.blocks:
                content.extend(
                    _transform_content(
                        content=block.content,
                        inline_latex=inline_latex,
                        ref_id_to_number=None,
                    )
                )
            if content:
                elements.append(
                    TextBlock(
                        kind=TextKind.HEADLINE
                        if layout.ref in TITLE_TAGS
                        else TextKind.BODY,
                        level=layout.level,
                        content=content,
                    )
                )

    chapter_refs = search_references_in_chapter(chapter)
    for ref in chapter_refs:
        footnotes.append(
            Footnote(
                id=ref_id_to_number.get(ref.id, 1),
                contents=list(
                    _convert_reference_to_footnote_contents(
                        ref=ref, inline_latex=inline_latex, assets_path=assets_path
                    )
                ),
            )
        )

    return ChapterRecord(elements=elements, footnotes=footnotes)


def _extract_text_from_content(
    content: list[str | InlineExpression | Reference | HTMLTag],
) -> str:
    """Extract plain text from asset content list (for Formula latex_expression)."""
    parts = []
    for item in flatten(content):
        if isinstance(item, str):
            parts.append(item)
        elif isinstance(item, InlineExpression):
            parts.append(item.content)
    return "".join(parts).strip()


def _convert_asset_to_epub(
    asset: AssetLayout,
    assets_path: Path,
    inline_latex: bool = False,
    ref_id_to_number: dict | None = None,
):
    title = list(
        _transform_content(
            content=asset.title,
            inline_latex=inline_latex,
            ref_id_to_number=ref_id_to_number,
        )
    )
    caption = list(
        _transform_content(
            content=asset.caption,
            inline_latex=inline_latex,
            ref_id_to_number=ref_id_to_number,
        )
    )
    if asset.ref == "equation":
        latex_expression = _extract_text_from_content(asset.content)
        if not latex_expression:
            return None

        return Formula(
            latex_expression=latex_expression,
            title=title,
            caption=caption,
        )

    elif asset.ref == "image":
        if asset.hash is None:
            return None

        image_file = assets_path / f"{asset.hash}.png"
        if not image_file.exists():
            return None

        return Image(
            path=image_file,
            title=title,
            caption=caption,
        )

    elif asset.ref == "table":
        if asset.hash is None:
            return None

        html_content: EpubHTMLTag | None = None
        for item in _transform_content(
            content=asset.content,
            inline_latex=inline_latex,
            ref_id_to_number=ref_id_to_number,
        ):
            if isinstance(item, EpubHTMLTag):
                html_content = item
                break

        if html_content is None:
            table_file = assets_path / f"{asset.hash}.png"
            if not table_file.exists():
                return None
            return Image(
                path=table_file,
                title=title,
                caption=caption,
            )
        else:
            return Table(
                title=title,
                caption=caption,
                html_content=html_content,
            )

    return None


def _convert_reference_to_footnote_contents(
    ref: Reference,
    assets_path: Path,
    inline_latex: bool,
):
    for layout in ref.layouts:
        if isinstance(layout, AssetLayout):
            asset_element = _convert_asset_to_epub(
                asset=layout,
                assets_path=assets_path,
                inline_latex=inline_latex,
                ref_id_to_number=None,
            )
            if asset_element:
                yield asset_element
        elif isinstance(layout, ParagraphLayout):
            content: list[str | Formula | Mark | EpubHTMLTag] = []
            for block in layout.blocks:
                content.extend(
                    _transform_content(
                        content=block.content,
                        inline_latex=inline_latex,
                        ref_id_to_number=None,
                    )
                )
            if content:
                yield TextBlock(
                    kind=TextKind.BODY,
                    level=layout.level,
                    content=content,
                )


def _transform_content(
    content: list[str | InlineExpression | Reference | HTMLTag],
    inline_latex: bool,
    ref_id_to_number: dict | None = None,
) -> Generator[str | Formula | Mark | EpubHTMLTag, None, None]:
    for item in content:
        if isinstance(item, str):
            yield item

        elif isinstance(item, InlineExpression):
            if inline_latex:
                yield Formula(latex_expression=item.content.strip())
            else:
                yield latex_to_plain_text(latex_content=item.content.strip())

        elif ref_id_to_number and isinstance(item, Reference):
            ref_number = ref_id_to_number.get(item.id, 1)
            yield Mark(id=ref_number)

        elif isinstance(item, HTMLTag):
            yield EpubHTMLTag(
                name=item.definition.name,
                attributes=item.attributes,
                content=list(
                    _transform_content(
                        content=item.children,
                        inline_latex=inline_latex,
                        ref_id_to_number=ref_id_to_number,
                    )
                ),
            )



import sys
from typing import Generic, TypeVar

P = TypeVar("P")


class _Group(Generic[P]):
    def __init__(self, items: list[tuple[float, P]]):
        self._items: list[tuple[float, P]] = items
        self._cached_cv: float | None = None
        self._cached_size: float | None = None

    @property
    def items(self) -> list[tuple[float, P]]:
        return self._items

    @property
    def cv(self) -> float:
        if self._cached_cv is None:
            self._cached_cv = self._calculate_cv(
                values=[size for size, _ in self._items],
            )
        return self._cached_cv

    @property
    def size(self) -> float:
        if self._cached_size is None:
            if self._items:
                self._cached_size = sum(size for size, _ in self._items) / len(
                    self._items
                )
            else:
                self._cached_size = 0.0
        return self._cached_size

    def _calculate_cv(self, values: list[float]) -> float:
        if not values or len(values) < 2:
            return 0.0
        mean = sum(values) / len(values)
        if mean == 0:
            return float("inf")
        variance = sum((x - mean) ** 2 for x in values) / len(values)
        std = variance**0.5
        return std / mean


def split_by_cv(
    payload_items: list[tuple[float, P]],
    max_cv: float = 0.0,
    max_groups: int = sys.maxsize,
) -> list[list[P]]:
    """通过控制 CV （变异系数）将 payload 分组，返回分组后的 payload 列表"""

    if len(payload_items) <= 2:
        return [[payload for _, payload in payload_items]]

    groups: list[_Group[P]] = [_Group(items=payload_items)]

    while len(groups) < max_groups:
        max_cv_group_index = _find_max_cv_group_index(groups, max_cv)
        if max_cv_group_index == -1:
            break

        splitted = _split_group_by_max_gap(groups[max_cv_group_index].items)
        if splitted is None:
            break

        group1_items, group2_items = splitted
        groups[max_cv_group_index] = _Group(items=group1_items)
        groups.insert(max_cv_group_index + 1, _Group(items=group2_items))

    return [
        [payload for _, payload in group.items]
        for group in sorted(groups, key=lambda g: g.size)
    ]


def _find_max_cv_group_index(
    groups: list[_Group[P]],
    max_cv: float,
) -> int:
    max_cv_group_index = -1
    max_cv_value = max_cv

    for i, group in enumerate(groups):
        if len(group.items) <= 2:
            continue

        if group.cv > max_cv_value:
            max_cv_value = group.cv
            max_cv_group_index = i

    return max_cv_group_index


def _split_group_by_max_gap(
    group: list[tuple[float, P]],
) -> tuple[list[tuple[float, P]], list[tuple[float, P]]] | None:
    sorted_items = sorted(group, key=lambda x: x[0])
    gaps: list[tuple[float, int]] = []

    for i in range(len(sorted_items) - 1):
        gap = sorted_items[i + 1][0] - sorted_items[i][0]
        gaps.append((gap, i))

    if not gaps:
        return None

    _, split_index = max(gaps, key=lambda x: x[0])
    group1 = sorted_items[: split_index + 1]
    group2 = sorted_items[split_index + 1 :]

    return group1, group2



from dataclasses import dataclass
from enum import Enum, auto
from typing import Generator, Generic, Iterable, TypeVar

from ..common import avg, split_by_cv
from ..pdf import PageLayout

_T = TypeVar("_T")
_CV = 0.1  # Coefficient of Variation

_MIN_SIZE_RATE = 0.15


@dataclass
class _Projection(Generic[_T]):
    center: float
    size: float
    weight: float
    payload: _T


def split_reading_serials(
    raw_layouts: list[PageLayout],
) -> Generator[list[PageLayout], None, None]:
    """
    将 OCR 识别的文字块按列分组，用于多列布局的阅读顺序识别。

    问题背景：
    - 扫描文档可能包含多列布局（双栏、三栏等学术论文或书籍）
    - 图片、表格等元素可能导致局部文字块被挤压，形成临时的列布局
    - 需要识别这些列的边界，以便按正确的阅读顺序（从左到右，从上到下）处理文字

    算法思路：
    1. 将所有 layout 在 x 轴上投影（使用中心点和宽度）
    2. 构建加权直方图（高度作为权重，避免小字符干扰）
    3. 分析直方图的波峰和波谷：波峰对应列中心，波谷对应列间隙
    4. 在显著的波谷处切分，将文字块分配到对应的列组
    5. 对每一列进一步切分，保证 CV （标准差 / 平均值）在合理范围，此步骤用于被图片挤压的文字块

    输入：原始的 layout 列表（从 OCR 获取，按页面位置无序）
    输出：按列分组的 layout 生成器，每组代表一列中的文字块
    """
    if not raw_layouts:
        return

    layout_pairs: list[tuple[int, int, PageLayout]] = []  # order, group_id, layout
    for group_id, group_layouts in enumerate(
        _group_projects(
            raw_projections=(
                _wrap_projection(order, layout)
                for order, layout in enumerate(raw_layouts)
            )
        )
    ):
        for order, layout in group_layouts:
            layout_pairs.append((order, group_id, layout))

    last_group_id = -1
    layouts_buffer: list[PageLayout] = []
    for _, group_id, layout in sorted(layout_pairs, key=lambda p: p[0]):
        if group_id != last_group_id:
            last_group_id = group_id
            if layouts_buffer:
                yield layouts_buffer
                layouts_buffer = []
        layouts_buffer.append(layout)

    if layouts_buffer:
        yield layouts_buffer


def _wrap_projection(
    index: int, layout: PageLayout
) -> _Projection[tuple[int, PageLayout]]:
    x1, y1, x2, y2 = layout.det
    return _Projection(
        center=(x1 + x2) / 2,
        size=x2 - x1,
        weight=float(y2 - y1),
        payload=(index, layout),
    )


def _group_projects(
    raw_projections: Iterable[_Projection[_T]],
) -> Generator[list[_T], None, None]:
    projections = list(raw_projections)
    avg_size = avg(p.size for p in projections)
    min_size_threshold = avg_size * _MIN_SIZE_RATE

    rectangles: list[_Rect] = []
    for p in projections:
        size = max(p.size, min_size_threshold)  # 避免毛刺
        rectangles.append(
            _Rect(left=p.center - size / 2, right=p.center + size / 2, height=p.weight)
        )

    for valley in _find_valleys(rectangles):
        next_group: list[_Projection] = []
        for projection in projections:
            if projection.center < valley:
                next_group.append(projection)
        for project in next_group:
            projections.remove(project)
        if next_group:
            for sub_group in split_by_cv(
                payload_items=[(p.size, p) for p in next_group],
                max_cv=_CV,
            ):
                yield [p.payload for p in sub_group]

    if projections:
        for sub_group in split_by_cv(
            payload_items=[(p.size, p) for p in projections],
            max_cv=_CV,
        ):
            yield [p.payload for p in sub_group]


@dataclass
class _Rect:
    left: float
    right: float
    height: float


def _find_valleys(rectangles: Iterable[_Rect]):
    window: list[tuple[float, float]] = []
    prev_class: _WindowClass = _WindowClass.OTHER
    flat_buffer: list[float] = []

    for rect in _histograms(rectangles):
        center = (rect.left + rect.right) / 2
        window.append((center, rect.height))
        if len(window) > 3:
            window.pop(0)
        if len(window) != 3:
            continue
        prev, curr, next = window  # pylint: disable=unbalanced-tuple-unpacking
        clazz = _classify_window(prev, curr, next)
        keep_buffer = False

        if clazz == _WindowClass.TOUCHED_GROUND:
            flat_buffer = [curr[0]]
            keep_buffer = True
        elif clazz == _WindowClass.LEFT_GROUND:
            if prev_class in (_WindowClass.TOUCHED_GROUND, _WindowClass.FLAT_GROUND):
                flat_buffer.append(curr[0])
                yield sum(flat_buffer) / len(flat_buffer)
        elif clazz == _WindowClass.FLAT_GROUND:
            if prev_class == _WindowClass.TOUCHED_GROUND or (
                prev_class == _WindowClass.FLAT_GROUND and flat_buffer
            ):
                flat_buffer.append(curr[0])
                keep_buffer = True
        elif clazz == _WindowClass.AT_VALLEY:
            yield curr[0]

        prev_class = clazz
        if not keep_buffer and flat_buffer:
            flat_buffer = []


def _histograms(raw_rectangles: Iterable[_Rect]):
    rectangles = list(raw_rectangles)
    rectangles.sort(key=lambda r: (r.left, r.right))
    forbidden: float = float("-inf")

    for i, rect in enumerate(rectangles):
        left = max(rect.left, forbidden)
        right = rect.right
        for j in range(i + 1, len(rectangles)):
            next_rect = rectangles[j]
            if next_rect.height > rect.height:
                right = min(right, next_rect.left)
        if left < right:
            yield _Rect(left=left, right=right, height=rect.height)
            forbidden = right


class _WindowClass(Enum):
    TOUCHED_GROUND = auto()
    LEFT_GROUND = auto()
    FLAT_GROUND = auto()
    AT_VALLEY = auto()
    OTHER = auto()


def _classify_window(
    prev: tuple[float, float],
    curr: tuple[float, float],
    next: tuple[float, float],
) -> _WindowClass:
    _, prev_h = prev
    _, curr_h = curr
    _, next_h = next
    if prev_h > curr_h and curr_h == next_h:
        return _WindowClass.TOUCHED_GROUND
    elif prev_h == curr_h and curr_h < next_h:
        return _WindowClass.LEFT_GROUND
    elif prev_h == curr_h and curr_h == next_h:
        return _WindowClass.FLAT_GROUND
    elif prev_h > curr_h < next_h:
        return _WindowClass.AT_VALLEY
    else:
        return _WindowClass.OTHER



import re
import unicodedata

from ..language import is_latin_letter

_LINK_FLAGS = frozenset(["‐", "‑", "‒", "–", "—", "―"])

# 全面的标点符号列表，包含多语言支持
# 参考资料：
# - https://en.wikipedia.org/wiki/General_Punctuation
# - https://www.unicode.org/charts/PDF/U2000.pdf
# - https://en.wikipedia.org/wiki/CJK_Symbols_and_Punctuation
# - https://www.unicode.org/charts/PDF/U3000.pdf
# - https://en.wiktionary.org/wiki/Appendix:Unicode/Supplemental_Punctuation
# - https://symbl.cc/en/unicode/blocks/supplemental-punctuation/
# - https://www.compart.com/en/unicode/category/Po
# - https://www.fileformat.info/info/unicode/category/Po/list.htm
# - https://en.wikipedia.org/wiki/Chinese_punctuation
# - https://en.wikipedia.org/wiki/Korean_punctuation
# - https://en.wikipedia.org/wiki/Guillemet
# - https://lingoculture.com/blog/grammar/french-punctuation-marks/
_PUNCTUATIONS = frozenset(
    [
        # 基本 ASCII 标点符号
        "!",
        '"',
        "#",
        "%",
        "&",
        "'",
        "(",
        ")",
        "*",
        ",",
        "-",
        ".",
        "/",
        ":",
        ";",
        "?",
        "@",
        "[",
        "\\",
        "]",
        "^",
        "_",
        "`",
        "{",
        "|",
        "}",
        "~",
        # 拉丁文补充区标点符号
        "¡",
        "§",
        "¶",
        "·",
        "¿",
        # 通用标点符号（U+2000-206F）
        "‐",
        "‑",
        "‒",
        "–",
        "—",
        "―",  # 各种连字符和破折号
        "‖",
        "‗",  # 双竖线和下划线
        "'",
        "'",
        "‚",
        "‛",  # 单引号变体
        """, """,
        "„",
        "‟",  # 双引号变体
        "†",
        "‡",  # 剑标
        "•",
        "‣",  # 项目符号
        "․",
        "‥",
        "…",  # 省略号
        "‧",  # 连字点
        "′",
        "″",
        "‴",
        "‵",
        "‶",
        "‷",  # 角分符号
        "‹",
        "›",  # 单角引号
        "※",  # 参考标记
        "‼",
        "‽",  # 双感叹号、疑问感叹号
        "‾",
        "‿",
        "⁀",
        "⁁",
        "⁂",
        "⁃",
        "⁄",
        "⁅",
        "⁆",
        "⁇",
        "⁈",
        "⁉",
        "⁊",
        "⁋",
        "⁌",
        "⁍",
        "⁎",
        "⁏",
        "⁐",
        "⁑",
        "⁒",
        "⁓",
        "⁔",
        "⁕",
        "⁖",
        "⁗",
        "⁘",
        "⁙",
        "⁚",
        "⁛",
        "⁜",
        "⁝",
        "⁞",
        # 补充标点符号（U+2E00-2E7F）
        "⸀",
        "⸁",
        "⸂",
        "⸃",
        "⸄",
        "⸅",
        "⸆",
        "⸇",
        "⸈",
        "⸉",
        "⸊",
        "⸋",
        "⸌",
        "⸍",
        "⸎",
        "⸏",
        "⸐",
        "⸑",
        "⸒",
        "⸓",
        "⸔",
        "⸕",
        "⸖",
        "⸗",
        "⸘",
        "⸙",
        "⸚",
        "⸛",
        "⸜",
        "⸝",
        "⸞",
        "⸟",
        "⸠",
        "⸡",
        "⸢",
        "⸣",
        "⸤",
        "⸥",
        "⸦",
        "⸧",
        "⸨",
        "⸩",
        "⸪",
        "⸫",
        "⸬",
        "⸭",
        "⸮",
        "ⸯ",
        "⸰",
        "⸱",
        "⸲",
        "⸳",
        "⸴",
        "⸵",
        "⸶",
        "⸷",
        "⸸",
        "⸹",
        "⸺",
        "⸻",
        "⸼",
        "⸽",
        "⸾",
        "⸿",
        "⹀",
        "⹁",
        "⹂",
        "⹃",
        "⹄",
        "⹅",
        "⹆",
        "⹇",
        "⹈",
        "⹉",
        "⹊",
        "⹋",
        "⹌",
        "⹍",
        "⹎",
        "⹏",
        # CJK 标点符号（U+3000-303F）
        "、",
        "。",  # 顿号、句号
        "〈",
        "〉",
        "《",
        "》",  # 书名号
        "「",
        "」",
        "『",
        "』",  # 角括号
        "【",
        "】",  # 方头括号
        "〔",
        "〕",  # 龟甲括号
        "〖",
        "〗",  # 白方头括号
        "〘",
        "〙",
        "〚",
        "〛",  # 其他括号
        "〜",
        "〝",
        "〞",
        "〟",  # 波浪号和引号
        "〰",
        "〽",  # 波浪线和乐谱号
        "・",  # 中点
        # 全角 ASCII 标点符号（U+FF00-FFEF）
        "！",
        "＂",
        "＃",
        "％",
        "＆",
        "＇",
        "（",
        "）",
        "＊",
        "，",
        "．",
        "／",
        "：",
        "；",
        "？",
        "＠",
        "［",
        "＼",
        "］",
        "＾",
        "＿",
        "｀",
        "｛",
        "｜",
        "｝",
        "～",
        "｡",
        "｢",
        "｣",
        "､",
        "･",  # 半角日文标点
        # 法语和德语特定标点符号
        "«",
        "»",  # 法语书名号（guillemets）
        "„",
        "‟",  # 德语引号
        # 西班牙语标点符号
        # "¡", "¿",  # 倒置的感叹号和问号（已在拉丁补充区中）
        # 希腊语标点符号
        "·",  # 希腊文中点（已包含）
        ";",  # 希腊问号（看起来像分号，已包含）
        # 阿拉伯语标点符号
        "؉",
        "؊",
        "،",
        "؍",
        "؎",
        "؏",
        "؛",
        "؞",
        "؟",
        "٪",
        "٫",
        "٬",
        "٭",
        # 希伯来语标点符号
        "֊",
        "־",
        "׀",
        "׃",
        "׆",
        "׳",
        "״",
        # 藏文标点符号
        "།",
        "༎",
        "༏",
        "༐",
        "༑",
        "༒",
        "༔",
        "༴",
        "༶",
        "༸",
        "྅",
        "࿐",
        "࿑",
        "࿒",
        "࿓",
        "࿔",
        "࿙",
        "࿚",
        # 缅甸语标点符号
        "၊",
        "။",
        "၌",
        "၍",
        "၎",
        "၏",
        # 其他亚洲文字标点符号
        "᙮",  # 加拿大音节句号
        "។",
        "៕",
        "៖",
        "៘",
        "៙",
        "៚",  # 高棉语标点
        "᠀",
        "᠁",
        "᠂",
        "᠃",
        "᠄",
        "᠅",
        "᠆",
        "᠇",
        "᠈",
        "᠉",
        "᠊",  # 蒙古语标点
        "჻",  # 格鲁吉亚标点
        "፠",
        "፡",
        "።",
        "፣",
        "፤",
        "፥",
        "፦",
        "፧",
        "፨",  # 埃塞俄比亚标点
        # 其他特殊标点符号
        "‱",  # 千分号
        "‸",  # 插入符号
        "※",  # 参考符号（已包含）
        "⁁",  # 插入点
        "⁓",  # 波浪连字符
    ]
)


def normalize_text(text: str) -> str:
    """
    扫描件中的文字杂乱，此方法尽可能规范化文字，以让相同语义的文字在字符串上也尽可能完全一致
    """
    text = re.sub(r"\s+", " ", text).strip()
    chars = _process_spaces_and_hyphens(text)
    return "".join(_remove_punctuation_and_normalize_latin(chars))


def _process_spaces_and_hyphens(text: str) -> list[str]:
    """
    针对拉丁字母语言相关的处理。检查连字符以拼回单词（连字符用于换行时被截断的单词）。
    然后删除非拉丁语字母语言之间的空格，对于汉语而言，删光字之间的空格不影响阅读。
    """
    chars: list[str] = []
    i = 0
    while i < len(text):
        char = text[i]
        if char == " ":
            # 规则1：检查是否是 拉丁字母 + 连字符 + 空格 + 拉丁字母
            if (
                len(chars) >= 2
                and chars[-1] in _LINK_FLAGS
                and is_latin_letter(chars[-2])
                and i < len(text) - 1
                and is_latin_letter(text[i + 1])
            ):
                # 删除连字符，跳过空格，直接连接单词
                chars.pop()
                i += 1
                continue

            # 规则2：只保留拉丁字母之间的空格
            keep_space = False
            if len(chars) > 0 and i < len(text) - 1:
                prev_is_latin = is_latin_letter(chars[-1])
                next_is_latin = is_latin_letter(text[i + 1])
                keep_space = prev_is_latin and next_is_latin

            if keep_space:
                chars.append(char)
            i += 1
        else:
            chars.append(char)
            i += 1

    return chars


def _remove_punctuation_and_normalize_latin(chars: list[str]):
    """
    进一步去除所有干扰因素。先删光标点符号，然后将拉丁字全传小写，最后删除重音符号。
    """
    for char in chars:
        if char in _PUNCTUATIONS:
            continue
        if not is_latin_letter(char):
            yield char
            continue
        char = char.lower()
        for d_char in unicodedata.normalize("NFD", char):
            # NFD 拆解以过滤重音符号
            if unicodedata.category(d_char) != "Mn":
                yield d_char



import re
from html import escape, unescape

from .tags import is_protocol_allowed, is_tag_filtered, is_tag_ignored, tag_definition
from .types import HTMLTag, P


def parse_raw_markdown(input: str) -> list[str | P | HTMLTag[P]]:
    """
    Parse raw markdown text containing HTML tags according to CommonMark and GFM specifications.

    This function:
    1. Removes dangerous HTML constructs (comments, processing instructions, declarations, CDATA)
    2. Filters tags and attributes according to GitHub's whitelist
    3. Escapes disallowed tags while exposing their children for recursive checking
    4. Applies GFM tagfilter to specific dangerous tags

    Args:
        input: Raw markdown text potentially containing HTML

    Returns:
        List of strings and HTMLTag objects representing the parsed content
    """
    result = []
    pos = 0

    while pos < len(input):
        # Look for the next "<"
        next_tag_pos = input.find("<", pos)

        if next_tag_pos == -1:
            # No more tags, add remaining text
            if pos < len(input):
                result.append(input[pos:])
            break

        # Add text before the tag
        if next_tag_pos > pos:
            result.append(input[pos:next_tag_pos])

        # Try to parse the HTML construct starting at next_tag_pos
        parsed, new_pos = _parse_html_construct(input, next_tag_pos)

        if parsed is not None:
            # Successfully parsed something
            if isinstance(parsed, list):
                result.extend(parsed)
            elif parsed:  # Skip empty strings
                result.append(parsed)
            pos = new_pos
        else:
            # Not a valid HTML construct, treat as literal text
            result.append("<")
            pos = next_tag_pos + 1

    return result


def _parse_html_construct(
    input: str, pos: int
) -> tuple[str | HTMLTag | list[str | HTMLTag] | None, int]:
    """
    Try to parse an HTML construct starting at position pos.

    Returns:
        Tuple of (parsed_result, new_position)
        - parsed_result can be None (no match), str, HTMLTag, or list
        - new_position is the position after the parsed construct
    """
    if not input[pos:].startswith("<"):
        return None, pos

    # Try to parse HTML comment
    if input[pos:].startswith("<!--"):
        end_pos = input.find("-->", pos + 4)
        if end_pos != -1:
            # Remove comment (GitHub removes these for security)
            return "", end_pos + 3
        return None, pos

    # Try to parse processing instruction
    if input[pos:].startswith("<?"):
        end_pos = input.find("?>", pos + 2)
        if end_pos != -1:
            # Remove processing instruction
            return "", end_pos + 2
        return None, pos

    # Try to parse CDATA section
    if input[pos:].startswith("<![CDATA["):
        end_pos = input.find("]]>", pos + 9)
        if end_pos != -1:
            # Remove CDATA section
            return "", end_pos + 3
        return None, pos

    # Try to parse declaration (<!DOCTYPE, etc.)
    if input[pos:].startswith("<!"):
        # Declaration must be followed by an ASCII letter
        if len(input) > pos + 2 and input[pos + 2].isalpha():
            end_pos = input.find(">", pos + 2)
            if end_pos != -1:
                # Remove declaration
                return "", end_pos + 1
        return None, pos

    # Try to parse a regular tag (opening, closing, or self-closing)
    return _parse_tag(input, pos)


def _parse_tag(
    input: str, pos: int
) -> tuple[str | HTMLTag[P] | list[str | P | HTMLTag[P]] | None, int]:
    """
    Parse an HTML tag (opening, closing, or self-closing).

    According to CommonMark spec:
    - Opening tag: <tagname attribute="value">
    - Closing tag: </tagname>
    - Self-closing tag: <tagname />
    """
    if not input[pos:].startswith("<"):
        return None, pos

    # Check if it's a closing tag
    is_closing = input[pos:].startswith("</")
    start_pos = pos + 2 if is_closing else pos + 1

    # Parse tag name
    # According to CommonMark: tag name consists of ASCII letters, digits, and hyphens
    tag_name_match = re.match(r"([a-zA-Z][a-zA-Z0-9-]*)", input[start_pos:])
    if not tag_name_match:
        return None, pos

    tag_name = tag_name_match.group(1).lower()
    pos_after_name = start_pos + len(tag_name_match.group(1))

    # For closing tags, just look for ">"
    if is_closing:
        # Skip optional whitespace
        ws_match = re.match(r"[ \t\n\r]*>", input[pos_after_name:])
        if ws_match:
            # Closing tags are just returned as escaped text for now
            # The actual tag matching logic would be handled by a higher-level parser
            end_pos = pos_after_name + len(ws_match.group(0))

            # Check if this tag should be filtered by GFM tagfilter
            if is_tag_filtered(tag_name):
                # Replace the leading "<" with "&lt;" to break the tag
                return "&lt;" + input[pos + 1 : end_pos], end_pos

            # Check if tag is in whitelist
            tag_def = tag_definition(tag_name)
            if tag_def:
                # Return the closing tag as-is (as text)
                return input[pos:end_pos], end_pos
            else:
                # Escape the closing tag
                return escape(input[pos:end_pos]), end_pos
        return None, pos

    # For opening tags, parse attributes
    attributes, pos_after_attrs, is_self_closing = _parse_attributes(
        input, pos_after_name
    )

    if pos_after_attrs is None:
        return None, pos

    # Check if this tag should be filtered by GFM tagfilter
    if is_tag_filtered(tag_name):
        # Replace the leading "<" with "&lt;" to break the tag
        return "&lt;" + input[pos + 1 : pos_after_attrs], pos_after_attrs

    # Check if this tag should be ignored (removed but children preserved)
    if is_tag_ignored(tag_name):
        # If self-closing, just remove it entirely
        if is_self_closing:
            return "", pos_after_attrs

        # For opening tags, find content and closing tag
        content, closing_tag_end = _parse_tag_content_and_closing(
            input, pos_after_attrs, tag_name
        )

        if content is not None:
            # Found closing tag, recursively parse content only (tags disappear)
            children = parse_raw_markdown(content) if content else []
            return children, closing_tag_end
        else:
            # No closing tag found, just remove the opening tag
            return "", closing_tag_end

    # Check if tag is in whitelist
    tag_def = tag_definition(tag_name)

    if tag_def:
        # Tag is allowed, filter attributes
        filtered_attrs = _filter_attributes(tag_def, attributes)

        # For self-closing tags, return HTMLTag with no children
        if is_self_closing:
            return HTMLTag(
                definition=tag_def, attributes=filtered_attrs, children=[]
            ), pos_after_attrs

        # For opening tags, find the closing tag and parse content
        content, closing_tag_end = _parse_tag_content_and_closing(
            input, pos_after_attrs, tag_name
        )

        if content is not None:
            # Found closing tag, recursively parse the content
            children: list[str | P | HTMLTag[P]] = (
                parse_raw_markdown(content) if content else []
            )
            return HTMLTag(
                definition=tag_def, attributes=filtered_attrs, children=children
            ), closing_tag_end
        else:
            # No closing tag found, treat as self-closing
            return HTMLTag(
                definition=tag_def, attributes=filtered_attrs, children=[]
            ), closing_tag_end
    else:
        # Tag is not allowed, escape the tag but expose children
        # Get the full tag text
        tag_text = input[pos:pos_after_attrs]

        # If it's a self-closing tag, just escape it
        if is_self_closing:
            return escape(tag_text), pos_after_attrs

        # For opening tags, we need to find the content and closing tag
        # Then escape the opening tag, recursively parse content, and escape closing tag
        content, closing_tag_end = _parse_tag_content_and_closing(
            input, pos_after_attrs, tag_name
        )

        if content is not None:
            # Found closing tag - escape opening and closing tags, but recursively parse content
            closing_tag = f"</{tag_name}>"
            result: list[str | P | HTMLTag[P]] = [escape(tag_text)]
            if content:
                result.extend(parse_raw_markdown(content))
            result.append(escape(closing_tag))
            return result, closing_tag_end
        else:
            # No closing tag found, just escape the opening tag
            return escape(tag_text), closing_tag_end


def _parse_tag_content_and_closing(
    input: str, content_start: int, tag_name: str
) -> tuple[str, int] | tuple[None, int]:
    """
    Find and extract content between opening and closing tags.

    Returns:
        Tuple of (content, closing_tag_end) if closing tag found
        Tuple of (None, content_start) if closing tag not found
    """
    closing_pos = _find_closing_tag(input, content_start, tag_name)

    if closing_pos != -1:
        # Found closing tag
        content = input[content_start:closing_pos]
        # Calculate the end position (skip past the closing tag)
        closing_tag_end = input.find(">", closing_pos)
        if closing_tag_end == -1:
            closing_tag_end = closing_pos + len(f"</{tag_name}>")
        else:
            closing_tag_end += 1
        return content, closing_tag_end
    else:
        # No closing tag found
        return None, content_start


def _parse_attributes(
    input: str, pos: int
) -> tuple[list[tuple[str, str]], int | None, bool]:
    """
    Parse HTML attributes from an opening tag.

    Returns:
        Tuple of (attributes, end_position, is_self_closing)
        - attributes: list of (name, value) tuples
        - end_position: position after the ">" or "/>", or None if parsing failed
        - is_self_closing: True if tag ends with "/>"
    """
    attributes = []
    current_pos = pos

    while current_pos < len(input):
        # Skip whitespace
        ws_match = re.match(r"[ \t\n\r]+", input[current_pos:])
        if ws_match:
            current_pos += len(ws_match.group(0))

        # Check for end of tag
        if input[current_pos:].startswith("/>"):
            return attributes, current_pos + 2, True
        if input[current_pos:].startswith(">"):
            return attributes, current_pos + 1, False

        # Parse attribute name
        # According to CommonMark: attribute name is [a-zA-Z_:][a-zA-Z0-9_.:-]*
        name_match = re.match(r"([a-zA-Z_:][a-zA-Z0-9_.:-]*)", input[current_pos:])
        if not name_match:
            # Invalid attribute, stop parsing
            break

        attr_name = name_match.group(1).lower()
        current_pos += len(name_match.group(1))

        # Skip whitespace after name
        ws_match = re.match(r"[ \t\n\r]*", input[current_pos:])
        if ws_match:
            current_pos += len(ws_match.group(0))

        # Check for "="
        if not input[current_pos:].startswith("="):
            # Attribute without value (boolean attribute)
            attributes.append((attr_name, ""))
            continue

        current_pos += 1

        # Skip whitespace after "="
        ws_match = re.match(r"[ \t\n\r]*", input[current_pos:])
        if ws_match:
            current_pos += len(ws_match.group(0))

        # Parse attribute value
        attr_value = ""
        if input[current_pos:].startswith('"'):
            # Double-quoted value
            current_pos += 1
            end_quote = input.find('"', current_pos)
            if end_quote == -1:
                # Unclosed quote, stop parsing
                break
            attr_value = input[current_pos:end_quote]
            current_pos = end_quote + 1
        elif input[current_pos:].startswith("'"):
            # Single-quoted value
            current_pos += 1
            end_quote = input.find("'", current_pos)
            if end_quote == -1:
                # Unclosed quote, stop parsing
                break
            attr_value = input[current_pos:end_quote]
            current_pos = end_quote + 1
        else:
            # Unquoted value
            # According to CommonMark: unquoted value is [^ \t\n\r"'=<>`]+
            value_match = re.match(r"([^ \t\n\r\"'=<>`]+)", input[current_pos:])
            if value_match:
                attr_value = value_match.group(1)
                current_pos += len(value_match.group(1))

        # Unescape HTML entities in attribute value
        attr_value = unescape(attr_value)

        attributes.append((attr_name, attr_value))

    # If we get here, we didn't find a proper tag ending
    return attributes, None, False


def _filter_attributes(
    tag_def, attributes: list[tuple[str, str]]
) -> list[tuple[str, str]]:
    """
    Filter attributes according to the tag's whitelist.

    Also validates URL protocols for href, src, and cite attributes.
    """
    filtered = []

    for name, value in attributes:
        # Check if attribute is allowed for this tag
        if name in tag_def.attributes:
            # For URL attributes, check protocol
            if name in ("href", "src", "cite"):
                if is_protocol_allowed(value):
                    filtered.append((name, value))
                # If protocol is not allowed, skip this attribute
            else:
                filtered.append((name, value))

    return filtered


def _find_closing_tag(input: str, start_pos: int, tag_name: str) -> int:
    """
    Find the matching closing tag for a given tag name.

    Handles nested tags of the same name and case-insensitive matching.
    """
    closing_tag_lower = f"</{tag_name.lower()}"
    opening_tag_lower = f"<{tag_name.lower()}"
    pos = start_pos
    depth = 1

    while pos < len(input):
        # Look for opening or closing tags of the same name (case-insensitive)
        next_open = -1
        next_close = -1

        # Search for next opening tag (case-insensitive)
        search_pos = pos
        while search_pos < len(input):
            found = input[search_pos:].lower().find(opening_tag_lower)
            if found == -1:
                break
            candidate_pos = search_pos + found
            # Verify it's a complete opening tag
            after_tag = candidate_pos + len(opening_tag_lower)
            if after_tag < len(input):
                next_char = input[after_tag]
                if next_char in " \t\n\r>/":
                    next_open = candidate_pos
                    break
            search_pos = candidate_pos + 1

        # Search for next closing tag (case-insensitive)
        search_pos = pos
        while search_pos < len(input):
            found = input[search_pos:].lower().find(closing_tag_lower)
            if found == -1:
                break
            candidate_pos = search_pos + found
            # Verify it's a complete closing tag
            after_tag = candidate_pos + len(closing_tag_lower)
            if after_tag < len(input):
                next_char = input[after_tag]
                if next_char in " \t\n\r>":
                    next_close = candidate_pos
                    break
            elif after_tag == len(input):
                # End of input, assume it needs '>'
                next_close = -1
                break
            search_pos = candidate_pos + 1

        # Process the nearest tag
        if next_close != -1 and (next_open == -1 or next_close < next_open):
            # Found closing tag
            depth -= 1
            if depth == 0:
                return next_close
            pos = next_close + len(closing_tag_lower)
        elif next_open != -1:
            # Found opening tag
            depth += 1
            pos = next_open + len(opening_tag_lower)
        else:
            # No more tags found
            break

    return -1



from dataclasses import dataclass
from typing import Callable, Generator, Generic, Iterable, TypeVar
from xml.etree.ElementTree import Element

from .tags import HTMLTagDefinition, tag_definition

P = TypeVar("P")


@dataclass
class HTMLTag(Generic[P]):
    definition: HTMLTagDefinition
    attributes: list[tuple[str, str]]
    children: list["str | P | HTMLTag[P]"]


def flatten(children: Iterable[str | P | HTMLTag[P]]) -> Generator[str | P, None, None]:
    for child in children:
        if isinstance(child, HTMLTag):
            yield from flatten(child.children)
        else:
            yield child


def decode(
    root: Element, decode_payload: Callable[[Element], P]
) -> list[str | P | HTMLTag[P]]:
    children: list[str | P | HTMLTag[P]] = []
    if root.text:
        children.append(root.text)

    for child in root:
        tag_def = tag_definition(child.tag)
        if tag_def is not None:
            attributes = list(child.attrib.items())
            children.append(
                HTMLTag(
                    definition=tag_def,
                    attributes=attributes,
                    children=decode(child, decode_payload),
                )
            )
        else:
            children.append(decode_payload(child))

        if child.tail:
            children.append(child.tail)

    return children


def encode(
    root: Element,
    children: list[str | P | HTMLTag[P]],
    encode_payload: Callable[[P], Element],
) -> None:
    if not children:
        return

    last_element: Element | None = None
    for child in children:
        if isinstance(child, str):
            if last_element is None:
                root.text = (root.text or "") + child
            else:
                last_element.tail = (last_element.tail or "") + child
        elif isinstance(child, HTMLTag):
            element = Element(child.definition.name, dict(child.attributes))
            encode(element, child.children, encode_payload)
            root.append(element)
            last_element = element
        else:
            element = encode_payload(child)
            root.append(element)
            last_element = element



import sys
import time
from dataclasses import dataclass
from enum import Enum, auto
from os import PathLike
from pathlib import Path
from threading import Lock
from typing import Callable, Container, Generator, TypeVar

from ..common import AssetHub, save_xml
from ..error import IgnoreOCRErrorsChecker, IgnorePDFErrorsChecker, OCRError, PDFError
from ..metering import AbortedCheck, check_aborted
from ..to_path import to_path
from .handler import DefaultPDFHandler, PDFHandler
from .page_extractor import Page, PageExtractorNode, PageLayout
from .page_ref import PageRefContext
from .types import DeepSeekOCRSize, PDFDocumentMetadata, encode


class OCREventKind(Enum):
    START = auto()
    IGNORE = auto()
    SKIP = auto()
    RENDERED = auto()
    COMPLETE = auto()
    FAILED = auto()


@dataclass
class OCREvent:
    kind: OCREventKind
    page_index: int
    total_pages: int
    cost_time_ms: int = 0
    input_tokens: int = 0
    output_tokens: int = 0
    error: Exception | None = None


class OCR:
    def __init__(
        self,
        model_path: PathLike | str | None,
        pdf_handler: PDFHandler | None,
        local_only: bool,
    ) -> None:
        self._pdf_handler = pdf_handler
        self._pdf_handler_lock = Lock()
        self._extractor = PageExtractorNode(
            model_path=to_path(model_path) if model_path is not None else None,
            local_only=local_only,
        )

    def predownload(self, revision: str | None) -> None:
        self._extractor.download_models(revision)

    def load_models(self) -> None:
        self._extractor.load_models()

    def metadata(self, pdf_path: Path) -> PDFDocumentMetadata:
        document = self._get_pdf_handler().open(pdf_path)
        try:
            return document.metadata()
        finally:
            document.close()

    def recognize(
        self,
        pdf_path: Path,
        asset_path: Path,
        ocr_path: Path,
        ocr_size: DeepSeekOCRSize = "gundam",
        dpi: int | None = None,
        max_page_image_file_size: int | None = None,
        includes_footnotes: bool = False,
        ignore_pdf_errors: IgnorePDFErrorsChecker = False,
        ignore_ocr_errors: IgnoreOCRErrorsChecker = False,
        plot_path: Path | None = None,
        cover_path: Path | None = None,
        aborted: AbortedCheck = lambda: False,
        page_indexes: Container[int] = range(1, sys.maxsize),
        max_tokens: int | None = None,
        max_output_tokens: int | None = None,
        device_number: int | None = None,
    ) -> Generator[OCREvent, None, None]:
        ocr_path.mkdir(parents=True, exist_ok=True)
        if plot_path is not None:
            plot_path.mkdir(parents=True, exist_ok=True)

        done_path = ocr_path / "done"
        did_ignore_any: bool = False
        if done_path.exists():
            return

        remain_tokens: int | None = max_tokens
        remain_output_tokens: int | None = max_output_tokens

        with PageRefContext(
            pdf_path=pdf_path,
            pdf_handler=self._get_pdf_handler(),
        ) as refs:
            pages_count = refs.pages_count
            asset_hub = AssetHub(asset_path)

            for ref in refs:
                check_aborted(aborted)
                start_time = time.perf_counter()
                yield OCREvent(
                    kind=OCREventKind.START,
                    page_index=ref.page_index,
                    total_pages=pages_count,
                )
                if ref.page_index not in page_indexes:
                    elapsed_ms = int((time.perf_counter() - start_time) * 1000)
                    did_ignore_any = True
                    yield OCREvent(
                        kind=OCREventKind.IGNORE,
                        page_index=ref.page_index,
                        total_pages=pages_count,
                        cost_time_ms=elapsed_ms,
                    )
                    continue

                filename = f"page_{ref.page_index}.xml"
                file_path = ocr_path / filename

                if file_path.exists():
                    elapsed_ms = int((time.perf_counter() - start_time) * 1000)
                    yield OCREvent(
                        kind=OCREventKind.SKIP,
                        page_index=ref.page_index,
                        total_pages=pages_count,
                        cost_time_ms=elapsed_ms,
                    )
                else:
                    from doc_page_extractor import TokenLimitError

                    if remain_tokens is not None and remain_tokens <= 0:
                        raise TokenLimitError()
                    if remain_output_tokens is not None and remain_output_tokens <= 0:
                        raise TokenLimitError()

                    page: Page | None = None
                    recognized_error: Exception | None = None

                    try:
                        image = ref.render(
                            dpi=dpi
                            if dpi is not None
                            else 300,  # DPI=300 for scanned page
                            max_image_file_size=max_page_image_file_size,
                        )
                        yield OCREvent(
                            kind=OCREventKind.RENDERED,
                            page_index=ref.page_index,
                            total_pages=pages_count,
                            cost_time_ms=int((time.perf_counter() - start_time) * 1000),
                            input_tokens=0,
                            output_tokens=0,
                        )
                        page = self._extractor.image2page(
                            image=image,
                            page_index=ref.page_index,
                            asset_hub=asset_hub,
                            ocr_size=ocr_size,
                            includes_footnotes=includes_footnotes,
                            includes_raw_image=(ref.page_index == 1),
                            plot_path=plot_path,
                            max_tokens=remain_tokens,
                            max_output_tokens=remain_output_tokens,
                            device_number=device_number,
                            aborted=aborted,
                        )
                    except PDFError as error:
                        if not _check_ignore_error(ignore_pdf_errors, error):
                            raise
                        recognized_error = error

                    except OCRError as error:
                        if not _check_ignore_error(ignore_ocr_errors, error):
                            raise
                        recognized_error = error

                    if page is None:
                        page = self._create_warn_page(
                            page_index=ref.page_index,
                            text=f"[[Page {ref.page_index} extraction failed due to PDF rendering error]]",
                        )

                    save_xml(encode(page), file_path)

                    if cover_path and page.image:
                        cover_path.parent.mkdir(parents=True, exist_ok=True)
                        page.image.save(cover_path, format="PNG")

                    yield OCREvent(
                        kind=OCREventKind.COMPLETE
                        if recognized_error is None
                        else OCREventKind.FAILED,
                        error=recognized_error,
                        page_index=ref.page_index,
                        total_pages=pages_count,
                        cost_time_ms=int((time.perf_counter() - start_time) * 1000),
                        input_tokens=page.input_tokens,
                        output_tokens=page.output_tokens,
                    )
                    if remain_tokens is not None:
                        remain_tokens -= page.input_tokens
                        remain_tokens -= page.output_tokens

                    if remain_output_tokens is not None:
                        remain_output_tokens -= page.output_tokens

        if not did_ignore_any:
            done_path.touch()

    def _get_pdf_handler(self) -> PDFHandler:
        if self._pdf_handler is not None:
            return self._pdf_handler

        with self._pdf_handler_lock:
            if self._pdf_handler is None:
                self._pdf_handler = DefaultPDFHandler()
            return self._pdf_handler

    def _create_warn_page(self, page_index: int, text: str) -> Page:
        page = Page(
            index=page_index,
            image=None,
            body_layouts=[],
            footnotes_layouts=[],
            input_tokens=0,
            output_tokens=0,
        )
        page.body_layouts.append(
            PageLayout(
                ref="text",
                det=(0, 0, 100, 100),
                text=text,
                hash=None,
                order=0,
            )
        )
        return page


_T = TypeVar("_T", bound=Exception)


def _check_ignore_error(check: bool | Callable[[_T], bool], error: _T) -> bool:
    if isinstance(check, bool):
        return check
    else:
        return check(error)



from pathlib import Path

from epub_generator import ChapterGetter, TocItem

from ..common import read_xml
from ..toc import Toc, decode


class TocCollection:
    def __init__(self, toc_path: Path | None) -> None:
        self._root: list[Toc]
        self._root_toc_items: list[TocItem] = []
        self._extra_toc_items: list[TocItem] = []
        self._id_to_toc_item: dict[int, TocItem] = {}
        self._having_body_toc_set: set[int] = set()

        if toc_path:
            self._root = decode(read_xml(toc_path)).content
        else:
            self._root = []

    @property
    def target(self) -> list[TocItem]:
        return self._root_toc_items + self._extra_toc_items

    def collect(
        self,
        toc_id: int,
        title: str,
        have_body: bool,
        get_chapter: ChapterGetter | None,
    ) -> None:
        toc_item: TocItem | None = None
        stack = self._find_raw_toc_item_stack(toc_id)

        if stack:
            current_toc_items = self._root_toc_items
            for raw_item in stack:
                toc_item = self._find_or_append_toc_item(raw_item.id, current_toc_items)
                current_toc_items = toc_item.children
            assert toc_item
            toc_item.title = title
            toc_item.get_chapter = get_chapter
        else:
            toc_item = TocItem(
                title=title,
                get_chapter=get_chapter,
            )
            self._extra_toc_items.append(toc_item)
            self._id_to_toc_item[toc_id] = toc_item

        if have_body:
            self._having_body_toc_set.add(id(toc_item))

    def normalize(self) -> "TocCollection":
        self._clean_no_content_items(self._root_toc_items)
        self._clean_no_content_items(self._extra_toc_items)
        return self

    def _find_raw_toc_item_stack(self, toc_id: int) -> list[Toc]:
        index: int = 0
        current_toc_items: list[Toc] = self._root
        stack: list[tuple[int, list[Toc], Toc]] = []
        while True:
            if index >= len(current_toc_items):
                if not stack:
                    break
                index, current_toc_items, _ = stack.pop()
                index += 1
            else:
                item = current_toc_items[index]
                stack.append((index, current_toc_items, item))
                if item.id == toc_id:
                    break
                current_toc_items = item.children
                index = 0

        return [item for _, _, item in stack]

    def _find_or_append_toc_item(self, id: int, toc_items: list[TocItem]) -> TocItem:
        toc_item = self._id_to_toc_item.get(id, None)
        if toc_item:
            if toc_item in toc_items:
                return toc_item
            else:
                raise RuntimeError(
                    f"TOC item with ID {id} already exists in another branch."
                )
        else:
            toc_item = TocItem(
                title="unknown",
                get_chapter=None,
            )
            toc_items.append(toc_item)
            self._id_to_toc_item[id] = toc_item
        return toc_item

    def _clean_no_content_items(self, toc_items: list[TocItem]):
        index = 0
        while index < len(toc_items):
            should_keep = True
            toc_item = toc_items[index]
            self._clean_no_content_items(toc_item.children)

            if id(toc_item) not in self._having_body_toc_set and not toc_item.children:
                # 有子节点同时自己也没有内容，变为仅存在于目录中的章节有助于更好的阅读体验
                toc_item.get_chapter = None
                should_keep = False

            if should_keep:
                index += 1
            else:
                toc_items.pop(index)



"""
GitHub Flavored Markdown (GFM) HTML Tags and Attributes Whitelist

This module defines the HTML tags and attributes allowed by GitHub when rendering
Markdown content. The whitelist is based on GitHub's html-pipeline sanitization
filter and GFM specification.

## Generation Principle

GitHub Flavored Markdown follows the CommonMark specification but adds additional
processing for security and consistency:

1. **Tagfilter Extension (GFM Spec)**: Certain dangerous tags are filtered by
   replacing the leading '<' with '&lt;'. These tags are chosen because they
   change how HTML is interpreted in a way unique to them.

2. **Sanitization Whitelist**: GitHub uses a whitelist-based HTML sanitizer that
   only allows specific tags and attributes. Any tags or attributes not in the
   whitelist are removed during rendering.

3. **Protocol Restrictions**: URLs in href, src, and cite attributes are restricted
   to http://, https://, and relative paths.

4. **Post-processing**: GitHub.com and GitHub Enterprise perform additional
   post-processing and sanitization after GFM is converted to HTML.

## Reference URLs

- GFM Specification: https://github.github.com/gfm/
- html-pipeline (Ruby): https://github.com/gjtorikian/html-pipeline/blob/main/lib/html_pipeline/sanitization_filter.rb
- Community Documentation: https://gist.github.com/seanh/13a93686bf4c2cb16e658b3cf96807f2
- GitHub Markup Issue: https://github.com/github/markup/issues/245
- CommonMark Spec (HTML Blocks): https://spec.commonmark.org/0.30/#html-blocks
- CommonMark Spec (Raw HTML): https://spec.commonmark.org/0.30/#raw-html

## Notes

- The whitelist is not officially documented by GitHub in a single authoritative
  source. This list is compiled from the html-pipeline library source code and
  community testing.
- GitHub may update the whitelist without notice. The actual whitelist used in
  production may differ from open-source implementations.
- For security reasons, most event handlers (onclick, onerror, etc.) and style
  tags are explicitly disallowed.

Last updated: 2025-12-15
"""

from dataclasses import dataclass


@dataclass
class HTMLTagDefinition:
    """Definition of an HTML tag with its allowed attributes and type."""

    name: str
    attributes: frozenset[str]
    is_block: bool


# ============================================================================
# Filtered Tags (GFM Tagfilter Extension)
# ============================================================================

# These tags are filtered by GFM's tagfilter extension. The leading '<' is
# replaced with '&lt;' to prevent them from being rendered as HTML.
# Reference: https://github.github.com/gfm/ (Section 6.11: Disallowed Raw HTML)

_FILTERED_TAGS = frozenset(
    [
        "title",  # Changes document title
        "textarea",  # Form input that interprets content differently
        "style",  # Can inject CSS that affects the entire page
        "xmp",  # Deprecated tag that displays content as preformatted text
        "iframe",  # Can embed external content (security risk)
        "noembed",  # Fallback for embed elements
        "noframes",  # Fallback for frames
        "script",  # Can execute JavaScript (XSS risk)
        "plaintext",  # Deprecated tag that treats rest of page as plain text
    ]
)


# ============================================================================
# Ignored Tags
# ============================================================================

# These tags are completely removed from the output, but their children content
# is preserved and recursively processed. The tags themselves disappear without
# being escaped.

_IGNORE_TAGS = frozenset(
    [
        "left",
        "center",
        "right",
    ]
)


# ============================================================================
# Universal Attributes
# ============================================================================

# These attributes are allowed on most HTML elements in the whitelist.
# Based on: https://github.com/gjtorikian/html-pipeline/blob/main/lib/html_pipeline/sanitization_filter.rb

UNIVERSAL_ATTRIBUTES = frozenset(
    [
        # Standard HTML attributes
        "abbr",
        "accept",
        "accept-charset",
        "accesskey",
        "action",
        "align",
        "alt",
        "axis",
        "border",
        "cellpadding",
        "cellspacing",
        "char",
        "charoff",
        "charset",
        "checked",
        "clear",
        "cols",
        "colspan",
        "color",
        "compact",
        "coords",
        "datetime",
        "dir",
        "disabled",
        "enctype",
        "for",
        "frame",
        "headers",
        "height",
        "hreflang",
        "hspace",
        "id",
        "ismap",
        "label",
        "lang",
        "longdesc",
        "maxlength",
        "media",
        "method",
        "multiple",
        "name",
        "nohref",
        "noshade",
        "nowrap",
        "open",
        "prompt",
        "readonly",
        "rel",
        "rev",
        "rows",
        "rowspan",
        "rules",
        "scope",
        "selected",
        "shape",
        "size",
        "span",
        "start",
        "summary",
        "tabindex",
        "target",
        "title",
        "type",
        "usemap",
        "valign",
        "value",
        "vspace",
        "width",
        # ARIA attributes for accessibility
        "aria-describedby",
        "aria-hidden",
        "aria-label",
        "aria-labelledby",
        "role",
        # Microdata attributes
        "itemprop",
        "itemscope",
        "itemtype",
    ]
)


# ============================================================================
# Element-Specific Attributes
# ============================================================================

# Additional attributes allowed for specific elements only

ELEMENT_SPECIFIC_ATTRIBUTES = {
    "a": frozenset(["href", "hreflang"]),
    "img": frozenset(["src", "longdesc", "loading", "alt"]),
    "source": frozenset(["src", "srcset", "type", "media"]),
    "picture": frozenset([]),
    "div": frozenset(["itemscope", "itemtype"]),
    "blockquote": frozenset(["cite"]),
    "del": frozenset(["cite", "datetime"]),
    "ins": frozenset(["cite", "datetime"]),
    "q": frozenset(["cite"]),
    "time": frozenset(["datetime"]),
    "video": frozenset(["src", "poster", "controls", "width", "height"]),
}


# ============================================================================
# Allowed Protocol Schemes
# ============================================================================

# URL schemes allowed in href, src, and cite attributes
ALLOWED_PROTOCOLS = frozenset(
    [
        "http",
        "https",
        "mailto",
        # Relative URLs are also allowed (no protocol)
    ]
)


# ============================================================================
# HTML Tag Definitions
# ============================================================================


# Structural elements
HTML_DIV = HTMLTagDefinition(
    name="div",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["div"],
    is_block=True,
)

HTML_P = HTMLTagDefinition(
    name="p",
    attributes=UNIVERSAL_ATTRIBUTES,
    is_block=True,
)

HTML_BLOCKQUOTE = HTMLTagDefinition(
    name="blockquote",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["blockquote"],
    is_block=True,
)

HTML_DETAILS = HTMLTagDefinition(
    name="details",
    attributes=UNIVERSAL_ATTRIBUTES | {"open"},
    is_block=True,
)

HTML_SUMMARY = HTMLTagDefinition(
    name="summary",
    attributes=UNIVERSAL_ATTRIBUTES,
    is_block=True,
)

HTML_FIGURE = HTMLTagDefinition(
    name="figure",
    attributes=UNIVERSAL_ATTRIBUTES,
    is_block=True,
)

HTML_FIGCAPTION = HTMLTagDefinition(
    name="figcaption",
    attributes=UNIVERSAL_ATTRIBUTES,
    is_block=True,
)


# Heading elements
HTML_H1 = HTMLTagDefinition(name="h1", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_H2 = HTMLTagDefinition(name="h2", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_H3 = HTMLTagDefinition(name="h3", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_H4 = HTMLTagDefinition(name="h4", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_H5 = HTMLTagDefinition(name="h5", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_H6 = HTMLTagDefinition(name="h6", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)


# Text formatting elements (inline)
HTML_B = HTMLTagDefinition(name="b", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)
HTML_I = HTMLTagDefinition(name="i", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)
HTML_STRONG = HTMLTagDefinition(
    name="strong", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_EM = HTMLTagDefinition(name="em", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)
HTML_SMALL = HTMLTagDefinition(
    name="small", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_MARK = HTMLTagDefinition(
    name="mark", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_S = HTMLTagDefinition(name="s", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)
HTML_STRIKE = HTMLTagDefinition(
    name="strike", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_ABBR = HTMLTagDefinition(
    name="abbr", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_CITE = HTMLTagDefinition(
    name="cite", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_DFN = HTMLTagDefinition(
    name="dfn", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_KBD = HTMLTagDefinition(
    name="kbd", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_SAMP = HTMLTagDefinition(
    name="samp", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_VAR = HTMLTagDefinition(
    name="var", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_CODE = HTMLTagDefinition(
    name="code", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_PRE = HTMLTagDefinition(name="pre", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_TT = HTMLTagDefinition(name="tt", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)
HTML_Q = HTMLTagDefinition(
    name="q",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["q"],
    is_block=False,
)
HTML_BDO = HTMLTagDefinition(
    name="bdo", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_INS = HTMLTagDefinition(
    name="ins",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["ins"],
    is_block=False,
)
HTML_DEL = HTMLTagDefinition(
    name="del",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["del"],
    is_block=False,
)
HTML_SUP = HTMLTagDefinition(
    name="sup", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_SUB = HTMLTagDefinition(
    name="sub", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_SPAN = HTMLTagDefinition(
    name="span", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)


# List elements
HTML_OL = HTMLTagDefinition(
    name="ol",
    attributes=UNIVERSAL_ATTRIBUTES | {"start", "reversed", "type"},
    is_block=True,
)
HTML_UL = HTMLTagDefinition(name="ul", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_LI = HTMLTagDefinition(
    name="li",
    attributes=UNIVERSAL_ATTRIBUTES | {"value"},
    is_block=True,
)
HTML_DL = HTMLTagDefinition(name="dl", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_DT = HTMLTagDefinition(name="dt", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_DD = HTMLTagDefinition(name="dd", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)


# Table elements
HTML_TABLE = HTMLTagDefinition(
    name="table",
    attributes=UNIVERSAL_ATTRIBUTES | {"summary"},
    is_block=True,
)
HTML_THEAD = HTMLTagDefinition(
    name="thead", attributes=UNIVERSAL_ATTRIBUTES, is_block=True
)
HTML_TBODY = HTMLTagDefinition(
    name="tbody", attributes=UNIVERSAL_ATTRIBUTES, is_block=True
)
HTML_TFOOT = HTMLTagDefinition(
    name="tfoot", attributes=UNIVERSAL_ATTRIBUTES, is_block=True
)
HTML_TR = HTMLTagDefinition(name="tr", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_TD = HTMLTagDefinition(
    name="td",
    attributes=UNIVERSAL_ATTRIBUTES | {"colspan", "rowspan", "headers"},
    is_block=True,
)
HTML_TH = HTMLTagDefinition(
    name="th",
    attributes=UNIVERSAL_ATTRIBUTES | {"colspan", "rowspan", "headers", "scope"},
    is_block=True,
)
HTML_CAPTION = HTMLTagDefinition(
    name="caption", attributes=UNIVERSAL_ATTRIBUTES, is_block=True
)


# Media elements
HTML_IMG = HTMLTagDefinition(
    name="img",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["img"],
    is_block=False,
)
HTML_PICTURE = HTMLTagDefinition(
    name="picture",
    attributes=UNIVERSAL_ATTRIBUTES,
    is_block=False,
)
HTML_SOURCE = HTMLTagDefinition(
    name="source",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["source"],
    is_block=False,
)
HTML_VIDEO = HTMLTagDefinition(
    name="video",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["video"],
    is_block=True,
)


# Link and semantic elements
HTML_A = HTMLTagDefinition(
    name="a",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["a"],
    is_block=False,
)
HTML_BR = HTMLTagDefinition(name="br", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)
HTML_HR = HTMLTagDefinition(name="hr", attributes=UNIVERSAL_ATTRIBUTES, is_block=True)
HTML_TIME = HTMLTagDefinition(
    name="time",
    attributes=UNIVERSAL_ATTRIBUTES | ELEMENT_SPECIFIC_ATTRIBUTES["time"],
    is_block=False,
)
HTML_WBR = HTMLTagDefinition(
    name="wbr", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)


# Ruby annotation elements (for East Asian typography)
HTML_RUBY = HTMLTagDefinition(
    name="ruby", attributes=UNIVERSAL_ATTRIBUTES, is_block=False
)
HTML_RT = HTMLTagDefinition(name="rt", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)
HTML_RP = HTMLTagDefinition(name="rp", attributes=UNIVERSAL_ATTRIBUTES, is_block=False)


# ============================================================================
# Collections
# ============================================================================

# Map of tag names to their definitions
# This is the single source of truth for allowed HTML tags
_TAG_DEFINITIONS = {
    "div": HTML_DIV,
    "p": HTML_P,
    "blockquote": HTML_BLOCKQUOTE,
    "details": HTML_DETAILS,
    "summary": HTML_SUMMARY,
    "figure": HTML_FIGURE,
    "figcaption": HTML_FIGCAPTION,
    "h1": HTML_H1,
    "h2": HTML_H2,
    "h3": HTML_H3,
    "h4": HTML_H4,
    "h5": HTML_H5,
    "h6": HTML_H6,
    "b": HTML_B,
    "i": HTML_I,
    "strong": HTML_STRONG,
    "em": HTML_EM,
    "small": HTML_SMALL,
    "mark": HTML_MARK,
    "s": HTML_S,
    "strike": HTML_STRIKE,
    "abbr": HTML_ABBR,
    "cite": HTML_CITE,
    "dfn": HTML_DFN,
    "kbd": HTML_KBD,
    "samp": HTML_SAMP,
    "var": HTML_VAR,
    "code": HTML_CODE,
    "pre": HTML_PRE,
    "tt": HTML_TT,
    "q": HTML_Q,
    "bdo": HTML_BDO,
    "ins": HTML_INS,
    "del": HTML_DEL,
    "sup": HTML_SUP,
    "sub": HTML_SUB,
    "span": HTML_SPAN,
    "ol": HTML_OL,
    "ul": HTML_UL,
    "li": HTML_LI,
    "dl": HTML_DL,
    "dt": HTML_DT,
    "dd": HTML_DD,
    "table": HTML_TABLE,
    "thead": HTML_THEAD,
    "tbody": HTML_TBODY,
    "tfoot": HTML_TFOOT,
    "tr": HTML_TR,
    "td": HTML_TD,
    "th": HTML_TH,
    "caption": HTML_CAPTION,
    "img": HTML_IMG,
    "picture": HTML_PICTURE,
    "source": HTML_SOURCE,
    "video": HTML_VIDEO,
    "a": HTML_A,
    "br": HTML_BR,
    "hr": HTML_HR,
    "time": HTML_TIME,
    "wbr": HTML_WBR,
    "ruby": HTML_RUBY,
    "rt": HTML_RT,
    "rp": HTML_RP,
}

# All allowed HTML tags (tag names only)
# Derived from TAG_DEFINITIONS to avoid duplication
ALLOWED_TAGS = frozenset(_TAG_DEFINITIONS.keys())


# ============================================================================
# Helper Functions
# ============================================================================


def tag_definition(tag_name: str) -> HTMLTagDefinition | None:
    return _TAG_DEFINITIONS.get(tag_name.lower(), None)


def is_tag_filtered(tag_name: str) -> bool:
    return tag_name.lower() in _FILTERED_TAGS


def is_tag_ignored(tag_name: str) -> bool:
    return tag_name.lower() in _IGNORE_TAGS


def is_protocol_allowed(url: str) -> bool:
    if not url:
        return True

    # Relative URLs are allowed
    if url.startswith("/") or url.startswith("./") or url.startswith("../"):
        return True

    # Check against allowed protocols
    url_lower = url.lower()
    for protocol in ALLOWED_PROTOCOLS:
        if url_lower.startswith(f"{protocol}:"):
            return True

    return False
